from string import ascii_letters
import textwrap

from fontTools.misc.testTools import getXML
from fontTools import subset
from fontTools.fontBuilder import FontBuilder
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.ttLib import TTFont, newTable
from fontTools.subset.svg import NAMESPACES, ranges

import pytest

etree = pytest.importorskip("lxml.etree")


@pytest.fixture
def empty_svg_font():
    glyph_order = [".notdef"] + list(ascii_letters)

    pen = TTGlyphPen(glyphSet=None)
    pen.moveTo((0, 0))
    pen.lineTo((0, 500))
    pen.lineTo((500, 500))
    pen.lineTo((500, 0))
    pen.closePath()
    glyph = pen.glyph()
    glyphs = {g: glyph for g in glyph_order}

    fb = FontBuilder(unitsPerEm=1024, isTTF=True)
    fb.setupGlyphOrder(glyph_order)
    fb.setupCharacterMap({ord(c): c for c in ascii_letters})
    fb.setupGlyf(glyphs)
    fb.setupHorizontalMetrics({g: (500, 0) for g in glyph_order})
    fb.setupHorizontalHeader()
    fb.setupOS2()
    fb.setupPost()
    fb.setupNameTable({"familyName": "TestSVG", "styleName": "Regular"})

    svg_table = newTable("SVG ")
    svg_table.docList = []
    fb.font["SVG "] = svg_table

    return fb.font


# 'simple' here means one svg document per glyph. The required 'id' attribute
# containing the 'glyphXXX' indices can be either on a child of the root <svg>
# or on the <svg> root itself, so we test with both.
# see https://github.com/fonttools/fonttools/issues/2548


def simple_svg_table_glyph_ids_on_children(empty_svg_font):
    font = empty_svg_font
    svg_docs = font["SVG "].docList
    for i in range(1, 11):
        svg = new_svg()
        etree.SubElement(svg, "path", {"id": f"glyph{i}", "d": f"M{i},{i}"})
        svg_docs.append((etree.tostring(svg).decode(), i, i))
    return font


def simple_svg_table_glyph_ids_on_roots(empty_svg_font):
    font = empty_svg_font
    svg_docs = font["SVG "].docList
    for i in range(1, 11):
        svg = new_svg(id=f"glyph{i}")
        etree.SubElement(svg, "path", {"d": f"M{i},{i}"})
        svg_docs.append((etree.tostring(svg).decode(), i, i))
    return font


def new_svg(**attrs):
    return etree.Element("svg", {"xmlns": NAMESPACES["svg"], **attrs})


def _lines(s):
    return textwrap.dedent(s).splitlines()


@pytest.mark.parametrize(
    "add_svg_table, gids, retain_gids, expected_xml",
    [
        # keep four glyphs in total, don't retain gids, which thus get remapped
        (
            simple_svg_table_glyph_ids_on_children,
            "2,4-6",
            False,
            _lines(
                """\
                <svgDoc endGlyphID="1" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph1" d="M2,2"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="2" startGlyphID="2">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph2" d="M4,4"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="3" startGlyphID="3">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph3" d="M5,5"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="4" startGlyphID="4">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph4" d="M6,6"/></svg>]]>
                </svgDoc>
                """
            ),
        ),
        # same as above but with glyph id attribute in the root <svg> element itself
        # https://github.com/fonttools/fonttools/issues/2548
        (
            simple_svg_table_glyph_ids_on_roots,
            "2,4-6",
            False,
            _lines(
                """\
                <svgDoc endGlyphID="1" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph1"><path d="M2,2"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="2" startGlyphID="2">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph2"><path d="M4,4"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="3" startGlyphID="3">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph3"><path d="M5,5"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="4" startGlyphID="4">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph4"><path d="M6,6"/></svg>]]>
                </svgDoc>
                """
            ),
        ),
        # same four glyphs, but we now retain gids
        (
            simple_svg_table_glyph_ids_on_children,
            "2,4-6",
            True,
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="2">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph2" d="M2,2"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="4" startGlyphID="4">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph4" d="M4,4"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="5" startGlyphID="5">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph5" d="M5,5"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="6" startGlyphID="6">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><path id="glyph6" d="M6,6"/></svg>]]>
                </svgDoc>
                """
            ),
        ),
        # retain gids like above but with glyph id attribute in the root <svg> element itself
        # https://github.com/fonttools/fonttools/issues/2548
        (
            simple_svg_table_glyph_ids_on_roots,
            "2,4-6",
            True,
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="2">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph2"><path d="M2,2"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="4" startGlyphID="4">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph4"><path d="M4,4"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="5" startGlyphID="5">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph5"><path d="M5,5"/></svg>]]>
                </svgDoc>
                <svgDoc endGlyphID="6" startGlyphID="6">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" id="glyph6"><path d="M6,6"/></svg>]]>
                </svgDoc>
                """
            ),
        ),
    ],
)
def test_subset_single_glyph_per_svg(
    empty_svg_font, add_svg_table, tmp_path, gids, retain_gids, expected_xml
):
    font = add_svg_table(empty_svg_font)

    svg_font_path = tmp_path / "TestSVG.ttf"
    font.save(svg_font_path)

    subset_path = svg_font_path.with_suffix(".subset.ttf")

    subset.main(
        [
            str(svg_font_path),
            f"--output-file={subset_path}",
            f"--gids={gids}",
            "--retain_gids" if retain_gids else "--no-retain_gids",
        ]
    )
    subset_font = TTFont(subset_path)

    assert getXML(subset_font["SVG "].toXML, subset_font) == expected_xml


# This contains a bunch of cross-references between glyphs, paths, gradients, etc.
# Note the path coordinates are completely made up and not meant to be rendered.
# We only care about the tree structure, not it's visual content.
COMPLEX_SVG = """\
<svg xmlns="http://www.w3.org/2000/svg"
     xmlns:xlink="http://www.w3.org/1999/xlink">
  <defs>
    <linearGradient id="lg1" x1="50" x2="50" y1="80" y2="80" gradientUnits="userSpaceOnUse">
      <stop stop-color="#A47B62" offset="0"/>
      <stop stop-color="#AD8264" offset="1.0"/>
    </linearGradient>
    <radialGradient id="rg2" cx="50" cy="50" r="10" gradientUnits="userSpaceOnUse">
      <stop stop-color="#A47B62" offset="0"/>
      <stop stop-color="#AD8264" offset="1.0"/>
    </radialGradient>
    <radialGradient id="rg3" xlink:href="#rg2" r="20"/>
    <radialGradient id="rg4" xlink:href="#rg3" cy="100"/>
    <path id="p1" d="M3,3"/>
    <clipPath id="c1">
      <circle cx="10" cy="10" r="1"/>
    </clipPath>
  </defs>
  <g id="glyph1">
    <g id="glyph2">
      <path d="M0,0"/>
    </g>
    <g>
      <path d="M1,1" fill="url(#lg1)"/>
      <path d="M2,2"/>
    </g>
  </g>
  <g id="glyph3">
    <use xlink:href="#p1"/>
  </g>
  <use id="glyph4" xlink:href="#glyph1" x="10"/>
  <use id="glyph5" xlink:href="#glyph2" y="-10"/>
  <g id="glyph6">
    <use xlink:href="#p1" transform="scale(2, 1)"/>
  </g>
  <g id="group1">
    <g id="glyph7">
      <path id="p2" d="M4,4"/>
    </g>
    <g id=".glyph7">
      <path d="M4,4"/>
    </g>
    <g id="glyph8">
      <g id=".glyph8">
        <path id="p3" d="M5,5"/>
        <path id="M6,6"/>
      </g>
      <path d="M7,7"/>
    </g>
    <g id="glyph9">
      <use xlink:href="#p2"/>
    </g>
    <g id="glyph10">
      <use xlink:href="#p3"/>
    </g>
  </g>
  <g id="glyph11">
    <path d="M7,7" fill="url(#rg4)"/>
  </g>
  <g id="glyph12">
    <path d="M7,7" style="fill:url(#lg1);stroke:red;clip-path:url(#c1)"/>
  </g>
</svg>
"""


@pytest.mark.parametrize(
    "subset_gids, expected_xml",
    [
        # we only keep gid=2, with 'glyph2' defined inside 'glyph1': 'glyph2'
        # is renamed 'glyph1' to match the new subset indices, and the old 'glyph1'
        # is kept (as it contains 'glyph2') but renamed '.glyph1' to avoid clash
        (
            "2",
            _lines(
                """\
                <svgDoc endGlyphID="1" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <g id=".glyph1">
                    <g id="glyph1">
                      <path d="M0,0"/>
                    </g>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        # we keep both gid 1 and 2: the glyph elements' ids stay as they are (only the
        # range endGlyphID change); a gradient is kept since it's referenced by glyph1
        (
            "1,2",
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <defs>
                    <linearGradient id="lg1" x1="50" x2="50" y1="80" y2="80" gradientUnits="userSpaceOnUse">
                      <stop stop-color="#A47B62" offset="0"/>
                      <stop stop-color="#AD8264" offset="1.0"/>
                    </linearGradient>
                  </defs>
                  <g id="glyph1">
                    <g id="glyph2">
                      <path d="M0,0"/>
                    </g>
                    <g>
                      <path d="M1,1" fill="url(#lg1)"/>
                      <path d="M2,2"/>
                    </g>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        (
            # both gid 3 and 6 refer (via <use xlink:href="#...") to path 'p1', which
            # is thus kept in <defs>; the glyph ids and range start/end are renumbered.
            "3,6",
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <defs>
                    <path id="p1" d="M3,3"/>
                  </defs>
                  <g id="glyph1">
                    <use xlink:href="#p1"/>
                  </g>
                  <g id="glyph2">
                    <use xlink:href="#p1" transform="scale(2, 1)"/>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        (
            # 'glyph4' uses the whole 'glyph1' element (translated); we keep the latter
            # renamed to avoid clashes with new gids
            "3-4",
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <defs>
                    <linearGradient id="lg1" x1="50" x2="50" y1="80" y2="80" gradientUnits="userSpaceOnUse">
                      <stop stop-color="#A47B62" offset="0"/>
                      <stop stop-color="#AD8264" offset="1.0"/>
                    </linearGradient>
                    <path id="p1" d="M3,3"/>
                  </defs>
                  <g id=".glyph1">
                    <g id=".glyph2">
                      <path d="M0,0"/>
                    </g>
                    <g>
                      <path d="M1,1" fill="url(#lg1)"/>
                      <path d="M2,2"/>
                    </g>
                  </g>
                  <g id="glyph1">
                    <use xlink:href="#p1"/>
                  </g>
                  <use id="glyph2" xlink:href="#.glyph1" x="10"/>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        (
            # 'glyph9' uses a path 'p2' defined inside 'glyph7', the latter is excluded
            # from our subset, thus gets renamed '.glyph7'; an unrelated element with
            # same id=".glyph7" doesn't clash because it was dropped.
            # Similarly 'glyph10' uses path 'p3' defined inside 'glyph8', also excluded
            # from subset and prefixed with '.'. But since an id=".glyph8" is already
            # used in the doc, we append a .{digit} suffix to disambiguate.
            "9,10",
            _lines(
                """\
                <svgDoc endGlyphID="2" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <g id="group1">
                    <g id=".glyph7">
                      <path id="p2" d="M4,4"/>
                    </g>
                    <g id=".glyph8.1">
                      <g id=".glyph8">
                        <path id="p3" d="M5,5"/>
                      </g>
                    </g>
                    <g id="glyph1">
                      <use xlink:href="#p2"/>
                    </g>
                    <g id="glyph2">
                      <use xlink:href="#p3"/>
                    </g>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        (
            # 'glyph11' uses gradient 'rg4' which inherits from 'rg3', which inherits
            # from 'rg2', etc.
            "11",
            _lines(
                """\
                <svgDoc endGlyphID="1" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <defs>
                    <radialGradient id="rg2" cx="50" cy="50" r="10" gradientUnits="userSpaceOnUse">
                      <stop stop-color="#A47B62" offset="0"/>
                      <stop stop-color="#AD8264" offset="1.0"/>
                    </radialGradient>
                    <radialGradient id="rg3" xlink:href="#rg2" r="20"/>
                    <radialGradient id="rg4" xlink:href="#rg3" cy="100"/>
                  </defs>
                  <g id="glyph1">
                    <path d="M7,7" fill="url(#rg4)"/>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
        (
            # 'glyph12' contains a style attribute with inline CSS declarations that
            # contains references to a gradient fill and a clipPath: we keep those
            "12",
            _lines(
                """\
                <svgDoc endGlyphID="1" startGlyphID="1">
                  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink">
                  <defs>
                    <linearGradient id="lg1" x1="50" x2="50" y1="80" y2="80" gradientUnits="userSpaceOnUse">
                      <stop stop-color="#A47B62" offset="0"/>
                      <stop stop-color="#AD8264" offset="1.0"/>
                    </linearGradient>
                    <clipPath id="c1">
                      <circle cx="10" cy="10" r="1"/>
                    </clipPath>
                  </defs>
                  <g id="glyph1">
                    <path d="M7,7" style="fill:url(#lg1);stroke:red;clip-path:url(#c1)"/>
                  </g>
                </svg>
                ]]>
                </svgDoc>
                """
            ),
        ),
    ],
)
def test_subset_svg_with_references(
    empty_svg_font, tmp_path, subset_gids, expected_xml
):
    font = empty_svg_font

    font["SVG "].docList.append((COMPLEX_SVG, 1, 12))
    svg_font_path = tmp_path / "TestSVG.ttf"
    font.save(svg_font_path)
    subset_path = svg_font_path.with_suffix(".subset.ttf")

    subset.main(
        [
            str(svg_font_path),
            f"--output-file={subset_path}",
            f"--gids={subset_gids}",
            "--pretty-svg",
        ]
    )
    subset_font = TTFont(subset_path)

    if expected_xml is not None:
        assert getXML(subset_font["SVG "].toXML, subset_font) == expected_xml
    else:
        assert "SVG " not in subset_font


def test_subset_svg_empty_table(empty_svg_font, tmp_path):
    font = empty_svg_font

    svg = new_svg()
    etree.SubElement(svg, "rect", {"id": "glyph1", "x": "1", "y": "2"})
    font["SVG "].docList.append((etree.tostring(svg).decode(), 1, 1))

    svg_font_path = tmp_path / "TestSVG.ttf"
    font.save(svg_font_path)
    subset_path = svg_font_path.with_suffix(".subset.ttf")

    # there's no gid=2 in SVG table, drop the empty table
    subset.main([str(svg_font_path), f"--output-file={subset_path}", f"--gids=2"])

    assert "SVG " not in TTFont(subset_path)


def test_subset_svg_missing_glyph(empty_svg_font, tmp_path):
    font = empty_svg_font

    svg = new_svg()
    etree.SubElement(svg, "rect", {"id": "glyph1", "x": "1", "y": "2"})
    font["SVG "].docList.append(
        (
            etree.tostring(svg).decode(),
            1,
            # the range endGlyphID=2 declares two glyphs however our svg contains
            # only one glyph element with id="glyph1", the "glyph2" one is absent.
            # Techically this would be invalid according to the OT-SVG spec.
            2,
        )
    )
    svg_font_path = tmp_path / "TestSVG.ttf"
    font.save(svg_font_path)
    subset_path = svg_font_path.with_suffix(".subset.ttf")

    # make sure we don't crash when we don't find the expected "glyph2" element
    subset.main([str(svg_font_path), f"--output-file={subset_path}", f"--gids=1"])

    subset_font = TTFont(subset_path)
    assert getXML(subset_font["SVG "].toXML, subset_font) == [
        '<svgDoc endGlyphID="1" startGlyphID="1">',
        '  <![CDATA[<svg xmlns="http://www.w3.org/2000/svg"><rect id="glyph1" x="1" y="2"/></svg>]]>',
        "</svgDoc>",
    ]

    # ignore the missing gid even if included in the subset; in this test case we
    # end up with an empty svg document--which is dropped, along with the empty table
    subset.main([str(svg_font_path), f"--output-file={subset_path}", f"--gids=2"])

    assert "SVG " not in TTFont(subset_path)


@pytest.mark.parametrize(
    "ints, expected_ranges",
    [
        ((), []),
        ((0,), [(0, 0)]),
        ((0, 1), [(0, 1)]),
        ((1, 1, 1, 1), [(1, 1)]),
        ((1, 3), [(1, 1), (3, 3)]),
        ((4, 2, 1, 3), [(1, 4)]),
        ((1, 2, 4, 5, 6, 9, 13, 14, 15), [(1, 2), (4, 6), (9, 9), (13, 15)]),
    ],
)
def test_ranges(ints, expected_ranges):
    assert list(ranges(ints)) == expected_ranges



from __future__ import annotations

import typing
from pathlib import PurePosixPath

from ._base import FS
from ._errors import DirectoryExpected, ResourceNotFound

if typing.TYPE_CHECKING:
    from collections.abc import Collection
    from typing import IO, Any

    from ._info import Info


class SubFS(FS):
    """Maps a sub-directory of another filesystem."""

    def __init__(self, parent: FS, sub_path: str):
        super().__init__()
        self._parent = parent
        self._prefix = PurePosixPath(sub_path).as_posix().rstrip("/")
        if not parent.exists(self._prefix):
            raise ResourceNotFound(f"No such file or directory: {sub_path!r}")
        elif not parent.isdir(self._prefix):
            raise DirectoryExpected(f"{sub_path!r} is not a directory")

    def delegate_fs(self):
        return self._parent

    def _full(self, rel: str) -> str:
        self.check()
        return f"{self._prefix}/{PurePosixPath(rel).as_posix()}".lstrip("/")

    def open(self, path: str, mode: str = "rb", **kwargs) -> IO[Any]:
        return self._parent.open(self._full(path), mode, **kwargs)

    def exists(self, path: str) -> bool:
        return self._parent.exists(self._full(path))

    def isdir(self, path: str) -> bool:
        return self._parent.isdir(self._full(path))

    def isfile(self, path: str) -> bool:
        return self._parent.isfile(self._full(path))

    def listdir(self, path: str) -> list[str]:
        return self._parent.listdir(self._full(path))

    def makedir(self, path: str, recreate: bool = False):
        return self._parent.makedir(self._full(path), recreate=recreate)

    def makedirs(self, path: str, recreate: bool = False):
        return self._parent.makedirs(self._full(path), recreate=recreate)

    def getinfo(self, path: str, namespaces: Collection[str] | None = None) -> Info:
        return self._parent.getinfo(self._full(path), namespaces=namespaces)

    def remove(self, path: str):
        return self._parent.remove(self._full(path))

    def removedir(self, path: str):
        return self._parent.removedir(self._full(path))

    def removetree(self, path: str):
        return self._parent.removetree(self._full(path))

    def movedir(self, src: str, dst: str, create: bool = False):
        self._parent.movedir(self._full(src), self._full(dst), create=create)

    def getsyspath(self, path: str) -> str:
        return self._parent.getsyspath(self._full(path))

    def readbytes(self, path: str) -> bytes:
        return self._parent.readbytes(self._full(path))

    def writebytes(self, path: str, data: bytes):
        self._parent.writebytes(self._full(path), data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._parent!r}, {self._prefix!r})"

    def __str__(self) -> str:
        return f"{self._parent}/{self._prefix}"


class ClosingSubFS(SubFS):
    """Like SubFS, but auto-closes the parent filesystem when closed."""

    def close(self):
        super().close()
        self._parent.close()



from fontTools.cffLib import PrivateDict
from fontTools.cffLib.specializer import stringToProgram
from fontTools.misc.testTools import getXML, parseXML
from fontTools.misc.psCharStrings import (
    T2CharString,
    encodeFloat,
    encodeFixed,
    read_fixed1616,
    read_realNumber,
)
from fontTools.pens.recordingPen import RecordingPen
import unittest


def hexenc(s):
    return " ".join("%02x" % x for x in s)


class T2CharStringTest(unittest.TestCase):
    @classmethod
    def stringToT2CharString(cls, string):
        return T2CharString(program=stringToProgram(string), private=PrivateDict())

    def test_calcBounds_empty(self):
        cs = self.stringToT2CharString("endchar")
        bounds = cs.calcBounds(None)
        self.assertEqual(bounds, None)

    def test_calcBounds_line(self):
        cs = self.stringToT2CharString(
            "100 100 rmoveto 40 10 rlineto -20 50 rlineto endchar"
        )
        bounds = cs.calcBounds(None)
        self.assertEqual(bounds, (100, 100, 140, 160))

    def test_calcBounds_curve(self):
        cs = self.stringToT2CharString(
            "100 100 rmoveto -50 -150 200 0 -50 150 rrcurveto endchar"
        )
        bounds = cs.calcBounds(None)
        self.assertEqual(bounds, (91.90524980688875, -12.5, 208.09475019311125, 100))

    def test_charstring_bytecode_optimization(self):
        cs = self.stringToT2CharString(
            "100.0 100 rmoveto -50.0 -150 200.5 0.0 -50 150 rrcurveto endchar"
        )
        cs.isCFF2 = False
        cs.private._isCFF2 = False
        cs.compile()
        cs.decompile()
        self.assertEqual(
            cs.program,
            [
                100,
                100,
                "rmoveto",
                -50,
                -150,
                200.5,
                0,
                -50,
                150,
                "rrcurveto",
                "endchar",
            ],
        )

        cs2 = self.stringToT2CharString(
            "100.0 rmoveto -50.0 -150 200.5 0.0 -50 150 rrcurveto"
        )
        cs2.isCFF2 = True
        cs2.private._isCFF2 = True
        cs2.compile(isCFF2=True)
        cs2.decompile()
        self.assertEqual(
            cs2.program, [100, "rmoveto", -50, -150, 200.5, 0, -50, 150, "rrcurveto"]
        )

    def test_encodeFloat(self):
        testNums = [
            # value                expected result
            (-9.399999999999999, "1e e9 a4 ff"),  # -9.4
            (9.399999999999999999, "1e 9a 4f"),  # 9.4
            (456.8, "1e 45 6a 8f"),  # 456.8
            (0.0, "1e 0f"),  # 0
            (-0.0, "1e 0f"),  # 0
            (1.0, "1e 1f"),  # 1
            (-1.0, "1e e1 ff"),  # -1
            (98765.37e2, "1e 98 76 53 7f"),  # 9876537
            (1234567890.0, "1e 12 34 56 79 b2 ff"),  # 12345679E2
            (9.876537e-4, "1e 98 76 53 7c 10 ff"),  # 9876537E-10
            (9.876537e4, "1e 98 76 5a 37 ff"),  # 98765.37
            (1000.0, "1e 1b 3f"),  # 1E3
            (-1000.0, "1e e1 b3 ff"),  # -1E3
            (1e8, "1e 1b 8f"),  # 1E8
            (1e-5, "1e 1c 5f"),  # 1E-5
            (1.2e8, "1e 12 b7 ff"),  # 12E7
            (1.2345e-5, "1e 12 34 5c 9f"),  # 12345E-9
            (9.0987654e8, "1e 90 98 76 54 0f"),  # 909876540
            (0.1, "1e a1 ff"),  # .1
            (-0.1, "1e ea 1f"),  # -.1
            (0.01, "1e 1c 2f"),  # 1e-2
            (-0.01, "1e e1 c2 ff"),  # -1e-2
            (0.0123, "1e 12 3c 4f"),  # 123e-4
            (-0.0123, "1e e1 23 c4 ff"),  # -123e-4
        ]

        for sample in testNums:
            encoded_result = encodeFloat(sample[0])

            # check to see if we got the expected bytes
            self.assertEqual(hexenc(encoded_result), sample[1])

            # check to see if we get the same value by decoding the data
            decoded_result = read_realNumber(
                None,
                None,
                encoded_result,
                1,
            )
            self.assertEqual(decoded_result[0], float("%.8g" % sample[0]))
            # We limit to 8 digits of precision to match the implementation
            # of encodeFloat.

    def test_encode_decode_fixed(self):
        testNums = [
            # value                expected hex      expected float
            (-9.399999999999999, "ff ff f6 99 9a", -9.3999939),
            (-9.4, "ff ff f6 99 9a", -9.3999939),
            (9.399999999999999999, "ff 00 09 66 66", 9.3999939),
            (9.4, "ff 00 09 66 66", 9.3999939),
            (456.8, "ff 01 c8 cc cd", 456.8000031),
            (-456.8, "ff fe 37 33 33", -456.8000031),
        ]

        for value, expected_hex, expected_float in testNums:
            encoded_result = encodeFixed(value)

            # check to see if we got the expected bytes
            self.assertEqual(hexenc(encoded_result), expected_hex)

            # check to see if we get the same value by decoding the data
            decoded_result = read_fixed1616(
                None,
                None,
                encoded_result,
                1,
            )
            self.assertAlmostEqual(decoded_result[0], expected_float)

    def test_toXML(self):
        program = [
            "107 53.4004 166.199 hstem",
            "174.6 163.801 vstem",
            "338.4 142.8 rmoveto",
            "28 0 21.9 9 15.8 18 15.8 18 7.9 20.79959 0 23.6 rrcurveto",
            "endchar",
        ]
        cs = self.stringToT2CharString(" ".join(program))

        self.assertEqual(getXML(cs.toXML), program)

    def test_fromXML(self):
        cs = T2CharString()
        for name, attrs, content in parseXML(
            [
                '<CharString name="period">' "  338.4 142.8 rmoveto",
                "  28 0 21.9 9 15.8 18 15.8 18 7.9 20.79959 0 23.6 rrcurveto",
                "  endchar" "</CharString>",
            ]
        ):
            cs.fromXML(name, attrs, content)

        expected_program = [
            338.3999939,
            142.8000031,
            "rmoveto",
            28,
            0,
            21.8999939,
            9,
            15.8000031,
            18,
            15.8000031,
            18,
            7.8999939,
            20.7995911,
            0,
            23.6000061,
            "rrcurveto",
            "endchar",
        ]

        self.assertEqual(len(cs.program), len(expected_program))
        for arg, expected_arg in zip(cs.program, expected_program):
            if isinstance(arg, str):
                self.assertIsInstance(expected_arg, str)
                self.assertEqual(arg, expected_arg)
            else:
                self.assertNotIsInstance(expected_arg, str)
                self.assertAlmostEqual(arg, expected_arg)

    def test_pen_closePath(self):
        # Test CFF2/T2 charstring: it does NOT end in "endchar"
        # https://github.com/fonttools/fonttools/issues/2455
        cs = self.stringToT2CharString(
            "100 100 rmoveto -50 -150 200 0 -50 150 rrcurveto"
        )
        pen = RecordingPen()
        cs.draw(pen)
        self.assertEqual(pen.value[-1], ("closePath", ()))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from pathlib import Path

import pytest
from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.designspaceLib.split import Range
from fontTools.ttLib import TTFont, newTable
from fontTools.varLib.stat import buildVFStatTable, getStatAxes, getStatLocations


@pytest.fixture
def datadir():
    return Path(__file__).parent / "../designspaceLib/data"


def test_getStatAxes(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")

    assert getStatAxes(
        doc, {"Italic": 0, "Width": Range(50, 150), "Weight": Range(200, 900)}
    ) == [
        {
            "values": [
                {
                    "flags": 0,
                    "name": {
                        "de": "Extraleicht",
                        "en": "Extra Light",
                        "fr": "Extra léger",
                    },
                    "nominalValue": 200.0,
                    "rangeMaxValue": 250.0,
                    "rangeMinValue": 200.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Light"},
                    "nominalValue": 300.0,
                    "rangeMaxValue": 350.0,
                    "rangeMinValue": 250.0,
                },
                {
                    "flags": 2,
                    "name": {"en": "Regular"},
                    "nominalValue": 400.0,
                    "rangeMaxValue": 450.0,
                    "rangeMinValue": 350.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Semi Bold"},
                    "nominalValue": 600.0,
                    "rangeMaxValue": 650.0,
                    "rangeMinValue": 450.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Bold"},
                    "nominalValue": 700.0,
                    "rangeMaxValue": 850.0,
                    "rangeMinValue": 650.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Black"},
                    "nominalValue": 900.0,
                    "rangeMaxValue": 900.0,
                    "rangeMinValue": 850.0,
                },
                {
                    "flags": 2,
                    "name": {"en": "Regular"},
                    "value": 400.0,
                    "linkedValue": 700.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Bold"},
                    "value": 700.0,
                    "linkedValue": 400.0,
                },
            ],
            "name": {"en": "Wéíght", "fa-IR": "قطر"},
            "ordering": 2,
            "tag": "wght",
        },
        {
            "values": [
                {"flags": 0, "name": {"en": "Condensed"}, "value": 50.0},
                {"flags": 3, "name": {"en": "Normal"}, "value": 100.0},
                {"flags": 0, "name": {"en": "Wide"}, "value": 125.0},
                {
                    "flags": 0,
                    "name": {"en": "Extra Wide"},
                    "nominalValue": 150.0,
                    "rangeMinValue": 150.0,
                },
            ],
            "name": {"en": "Width", "fr": "Chasse"},
            "ordering": 1,
            "tag": "wdth",
        },
        {
            "values": [
                {"flags": 2, "linkedValue": 1.0, "name": {"en": "Roman"}, "value": 0.0},
            ],
            "name": {"en": "Italic"},
            "ordering": 3,
            "tag": "ital",
        },
    ]

    assert getStatAxes(doc, {"Italic": 1, "Width": 100, "Weight": Range(400, 700)}) == [
        {
            "values": [
                {
                    "flags": 2,
                    "name": {"en": "Regular"},
                    "nominalValue": 400.0,
                    "rangeMaxValue": 450.0,
                    "rangeMinValue": 350.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Semi Bold"},
                    "nominalValue": 600.0,
                    "rangeMaxValue": 650.0,
                    "rangeMinValue": 450.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Bold"},
                    "nominalValue": 700.0,
                    "rangeMaxValue": 850.0,
                    "rangeMinValue": 650.0,
                },
                {
                    "flags": 2,
                    "name": {"en": "Regular"},
                    "value": 400.0,
                    "linkedValue": 700.0,
                },
                {
                    "flags": 0,
                    "name": {"en": "Bold"},
                    "value": 700.0,
                    "linkedValue": 400.0,
                },
            ],
            "name": {"en": "Wéíght", "fa-IR": "قطر"},
            "ordering": 2,
            "tag": "wght",
        },
        {
            "values": [
                {"flags": 3, "name": {"en": "Normal"}, "value": 100.0},
            ],
            "name": {"en": "Width", "fr": "Chasse"},
            "ordering": 1,
            "tag": "wdth",
        },
        {
            "values": [
                {"flags": 0, "name": {"en": "Italic"}, "value": 1.0},
            ],
            "name": {"en": "Italic"},
            "ordering": 3,
            "tag": "ital",
        },
    ]


def test_getStatLocations(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")

    assert getStatLocations(
        doc, {"Italic": 0, "Width": Range(50, 150), "Weight": Range(200, 900)}
    ) == [
        {
            "flags": 0,
            "location": {"ital": 0.0, "wdth": 50.0, "wght": 300.0},
            "name": {"en": "Some Style", "fr": "Un Style"},
        },
    ]
    assert getStatLocations(
        doc, {"Italic": 1, "Width": Range(50, 150), "Weight": Range(200, 900)}
    ) == [
        {
            "flags": 0,
            "location": {"ital": 1.0, "wdth": 100.0, "wght": 700.0},
            "name": {"en": "Other"},
        },
    ]


@pytest.mark.parametrize(
    "with_mac_names",
    [
        pytest.param(True, id="with_mac_names"),
        pytest.param(False, id="without_mac_names"),
    ],
)
def test_buildVFStatTable(datadir, with_mac_names):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")
    ttFont = TTFont()

    nameTable = newTable("name")
    nameTable.names = []
    ttFont["name"] = nameTable

    if with_mac_names:
        # addName adds a name string for both Macintosh and Windows platforms by default
        nameTable.addName("Regular")

    buildVFStatTable(ttFont, doc, vfName="Test_WghtWdth")

    assert "STAT" in ttFont

    name_recs = ttFont["name"].names
    assert len({nr.nameID for nr in name_recs}) == 15

    # test that mac names don't get added if there weren't any before
    mac_recs = [nr for nr in name_recs if nr.platformID == 1]
    if with_mac_names:
        assert len(mac_recs) > 1
    else:
        assert len(mac_recs) == 0



from fontTools.misc.fixedTools import (
    fixedToFloat,
    floatToFixed,
    floatToFixedToStr,
    fixedToStr,
    strToFixed,
    strToFixedToFloat,
)
import unittest


class FixedToolsTest(unittest.TestCase):
    def test_roundtrip(self):
        for bits in range(0, 15):
            for value in range(-(2 ** (bits + 1)), 2 ** (bits + 1)):
                self.assertEqual(value, floatToFixed(fixedToFloat(value, bits), bits))

    def test_fixedToFloat_precision14(self):
        self.assertAlmostEqual(0.7999878, fixedToFloat(13107, 14))
        self.assertEqual(0.0, fixedToFloat(0, 14))
        self.assertEqual(1.0, fixedToFloat(16384, 14))
        self.assertEqual(-1.0, fixedToFloat(-16384, 14))
        self.assertAlmostEqual(0.999939, fixedToFloat(16383, 14))
        self.assertAlmostEqual(-0.999939, fixedToFloat(-16383, 14))

    def test_fixedToFloat_precision6(self):
        self.assertAlmostEqual(-9.984375, fixedToFloat(-639, 6))
        self.assertAlmostEqual(-10.0, fixedToFloat(-640, 6))
        self.assertAlmostEqual(9.984375, fixedToFloat(639, 6))
        self.assertAlmostEqual(10.0, fixedToFloat(640, 6))

    def test_fixedToStr_precision14(self):
        self.assertEqual("0.8", fixedToStr(13107, 14))
        self.assertEqual("0.0", fixedToStr(0, 14))
        self.assertEqual("1.0", fixedToStr(16384, 14))
        self.assertEqual("-1.0", fixedToStr(-16384, 14))
        self.assertEqual("0.99994", fixedToStr(16383, 14))
        self.assertEqual("-0.99994", fixedToStr(-16383, 14))

    def test_fixedToStr_precision6(self):
        self.assertAlmostEqual("-9.98", fixedToStr(-639, 6))
        self.assertAlmostEqual("-10.0", fixedToStr(-640, 6))
        self.assertAlmostEqual("9.98", fixedToStr(639, 6))
        self.assertAlmostEqual("10.0", fixedToStr(640, 6))

    def test_floatToFixed_precision14(self):
        self.assertEqual(13107, floatToFixed(0.8, 14))
        self.assertEqual(16384, floatToFixed(1.0, 14))
        self.assertEqual(16384, floatToFixed(1, 14))
        self.assertEqual(-16384, floatToFixed(-1.0, 14))
        self.assertEqual(-16384, floatToFixed(-1, 14))
        self.assertEqual(0, floatToFixed(0, 14))

    def test_strToFixed_precision14(self):
        self.assertEqual(13107, strToFixed("0.8", 14))
        self.assertEqual(16384, strToFixed("1.0", 14))
        self.assertEqual(16384, strToFixed("1", 14))
        self.assertEqual(-16384, strToFixed("-1.0", 14))
        self.assertEqual(-16384, strToFixed("-1", 14))
        self.assertEqual(0, strToFixed("0", 14))

    def test_strToFixedToFloat_precision14(self):
        self.assertAlmostEqual(0.7999878, strToFixedToFloat("0.8", 14))
        self.assertEqual(0.0, strToFixedToFloat("0", 14))
        self.assertEqual(1.0, strToFixedToFloat("1.0", 14))
        self.assertEqual(-1.0, strToFixedToFloat("-1.0", 14))
        self.assertAlmostEqual(0.999939, strToFixedToFloat("0.99994", 14))
        self.assertAlmostEqual(-0.999939, strToFixedToFloat("-0.99994", 14))

    def test_floatToFixedToStr_precision14(self):
        self.assertEqual("0.8", floatToFixedToStr(0.7999878, 14))
        self.assertEqual("1.0", floatToFixedToStr(1.0, 14))
        self.assertEqual("1.0", floatToFixedToStr(1, 14))
        self.assertEqual("-1.0", floatToFixedToStr(-1.0, 14))
        self.assertEqual("-1.0", floatToFixedToStr(-1, 14))
        self.assertEqual("0.0", floatToFixedToStr(0, 14))

    def test_fixedToFloat_return_float(self):
        value = fixedToFloat(16384, 14)
        self.assertIsInstance(value, float)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



"""Visualize DesignSpaceDocument and resulting VariationModel."""

from fontTools.varLib.models import VariationModel, supportScalar
from fontTools.designspaceLib import DesignSpaceDocument
from matplotlib import pyplot
from mpl_toolkits.mplot3d import axes3d
from itertools import cycle
import math
import logging
import sys

log = logging.getLogger(__name__)


def stops(support, count=10):
    a, b, c = support

    return (
        [a + (b - a) * i / count for i in range(count)]
        + [b + (c - b) * i / count for i in range(count)]
        + [c]
    )


def _plotLocationsDots(locations, axes, subplot, **kwargs):
    for loc, color in zip(locations, cycle(pyplot.cm.Set1.colors)):
        if len(axes) == 1:
            subplot.plot([loc.get(axes[0], 0)], [1.0], "o", color=color, **kwargs)
        elif len(axes) == 2:
            subplot.plot(
                [loc.get(axes[0], 0)],
                [loc.get(axes[1], 0)],
                [1.0],
                "o",
                color=color,
                **kwargs,
            )
        else:
            raise AssertionError(len(axes))


def plotLocations(locations, fig, names=None, **kwargs):
    n = len(locations)
    cols = math.ceil(n**0.5)
    rows = math.ceil(n / cols)

    if names is None:
        names = [None] * len(locations)

    model = VariationModel(locations)
    names = [names[model.reverseMapping[i]] for i in range(len(names))]

    axes = sorted(locations[0].keys())
    if len(axes) == 1:
        _plotLocations2D(model, axes[0], fig, cols, rows, names=names, **kwargs)
    elif len(axes) == 2:
        _plotLocations3D(model, axes, fig, cols, rows, names=names, **kwargs)
    else:
        raise ValueError("Only 1 or 2 axes are supported")


def _plotLocations2D(model, axis, fig, cols, rows, names, **kwargs):
    subplot = fig.add_subplot(111)
    for i, (support, color, name) in enumerate(
        zip(model.supports, cycle(pyplot.cm.Set1.colors), cycle(names))
    ):
        if name is not None:
            subplot.set_title(name)
        subplot.set_xlabel(axis)
        pyplot.xlim(-1.0, +1.0)

        Xs = support.get(axis, (-1.0, 0.0, +1.0))
        X, Y = [], []
        for x in stops(Xs):
            y = supportScalar({axis: x}, support)
            X.append(x)
            Y.append(y)
        subplot.plot(X, Y, color=color, **kwargs)

        _plotLocationsDots(model.locations, [axis], subplot)


def _plotLocations3D(model, axes, fig, rows, cols, names, **kwargs):
    ax1, ax2 = axes

    axis3D = fig.add_subplot(111, projection="3d")
    for i, (support, color, name) in enumerate(
        zip(model.supports, cycle(pyplot.cm.Set1.colors), cycle(names))
    ):
        if name is not None:
            axis3D.set_title(name)
        axis3D.set_xlabel(ax1)
        axis3D.set_ylabel(ax2)
        pyplot.xlim(-1.0, +1.0)
        pyplot.ylim(-1.0, +1.0)

        Xs = support.get(ax1, (-1.0, 0.0, +1.0))
        Ys = support.get(ax2, (-1.0, 0.0, +1.0))
        for x in stops(Xs):
            X, Y, Z = [], [], []
            for y in Ys:
                z = supportScalar({ax1: x, ax2: y}, support)
                X.append(x)
                Y.append(y)
                Z.append(z)
            axis3D.plot(X, Y, Z, color=color, **kwargs)
        for y in stops(Ys):
            X, Y, Z = [], [], []
            for x in Xs:
                z = supportScalar({ax1: x, ax2: y}, support)
                X.append(x)
                Y.append(y)
                Z.append(z)
            axis3D.plot(X, Y, Z, color=color, **kwargs)

        _plotLocationsDots(model.locations, [ax1, ax2], axis3D)


def plotDocument(doc, fig, **kwargs):
    doc.normalize()
    locations = [s.location for s in doc.sources]
    names = [s.name for s in doc.sources]
    plotLocations(locations, fig, names, **kwargs)


def _plotModelFromMasters2D(model, masterValues, fig, **kwargs):
    assert len(model.axisOrder) == 1
    axis = model.axisOrder[0]

    axis_min = min(loc.get(axis, 0) for loc in model.locations)
    axis_max = max(loc.get(axis, 0) for loc in model.locations)

    import numpy as np

    X = np.arange(axis_min, axis_max, (axis_max - axis_min) / 100)
    Y = []

    for x in X:
        loc = {axis: x}
        v = model.interpolateFromMasters(loc, masterValues)
        Y.append(v)

    subplot = fig.add_subplot(111)
    subplot.plot(X, Y, "-", **kwargs)


def _plotModelFromMasters3D(model, masterValues, fig, **kwargs):
    assert len(model.axisOrder) == 2
    axis1, axis2 = model.axisOrder[0], model.axisOrder[1]

    axis1_min = min(loc.get(axis1, 0) for loc in model.locations)
    axis1_max = max(loc.get(axis1, 0) for loc in model.locations)
    axis2_min = min(loc.get(axis2, 0) for loc in model.locations)
    axis2_max = max(loc.get(axis2, 0) for loc in model.locations)

    import numpy as np

    X = np.arange(axis1_min, axis1_max, (axis1_max - axis1_min) / 100)
    Y = np.arange(axis2_min, axis2_max, (axis2_max - axis2_min) / 100)
    X, Y = np.meshgrid(X, Y)
    Z = []

    for row_x, row_y in zip(X, Y):
        z_row = []
        Z.append(z_row)
        for x, y in zip(row_x, row_y):
            loc = {axis1: x, axis2: y}
            v = model.interpolateFromMasters(loc, masterValues)
            z_row.append(v)
    Z = np.array(Z)

    axis3D = fig.add_subplot(111, projection="3d")
    axis3D.plot_surface(X, Y, Z, **kwargs)


def plotModelFromMasters(model, masterValues, fig, **kwargs):
    """Plot a variation model and set of master values corresponding
    to the locations to the model into a pyplot figure.  Variation
    model must have axisOrder of size 1 or 2."""
    if len(model.axisOrder) == 1:
        _plotModelFromMasters2D(model, masterValues, fig, **kwargs)
    elif len(model.axisOrder) == 2:
        _plotModelFromMasters3D(model, masterValues, fig, **kwargs)
    else:
        raise ValueError("Only 1 or 2 axes are supported")


def main(args=None):
    from fontTools import configLogger

    if args is None:
        args = sys.argv[1:]

    # configure the library logger (for >= WARNING)
    configLogger()
    # comment this out to enable debug messages from logger
    # log.setLevel(logging.DEBUG)

    if len(args) < 1:
        print("usage: fonttools varLib.plot source.designspace", file=sys.stderr)
        print("  or")
        print("usage: fonttools varLib.plot location1 location2 ...", file=sys.stderr)
        print("  or")
        print(
            "usage: fonttools varLib.plot location1=value1 location2=value2 ...",
            file=sys.stderr,
        )
        sys.exit(1)

    fig = pyplot.figure()
    fig.set_tight_layout(True)

    if len(args) == 1 and args[0].endswith(".designspace"):
        doc = DesignSpaceDocument()
        doc.read(args[0])
        plotDocument(doc, fig)
    else:
        axes = [chr(c) for c in range(ord("A"), ord("Z") + 1)]
        if "=" not in args[0]:
            locs = [dict(zip(axes, (float(v) for v in s.split(",")))) for s in args]
            plotLocations(locs, fig)
        else:
            locations = []
            masterValues = []
            for arg in args:
                loc, v = arg.split("=")
                locations.append(dict(zip(axes, (float(v) for v in loc.split(",")))))
                masterValues.append(float(v))
            model = VariationModel(locations, axes[: len(locations[0])])
            plotModelFromMasters(model, masterValues, fig)

    pyplot.show()


if __name__ == "__main__":
    import sys

    sys.exit(main())



import logging
from types import MappingProxyType

from fontTools.pens.basePen import MissingComponentError
from fontTools.pens.filterPen import (
    DecomposingFilterPen,
    DecomposingFilterPointPen,
    FilterPointPen,
    OnCurveFirstPointPen,
)
from fontTools.pens.pointPen import PointToSegmentPen, ReverseFlipped
from fontTools.pens.recordingPen import RecordingPen, RecordingPointPen

import pytest


class SimpleGlyph:
    def draw(self, pen):
        pen.moveTo((0, 0))
        pen.curveTo((1, 1), (2, 2), (3, 3))
        pen.closePath()

    def drawPoints(self, pen):
        pen.beginPath()
        pen.addPoint((0, 0), "line")
        pen.addPoint((1, 1))
        pen.addPoint((2, 2))
        pen.addPoint((3, 3), "curve")
        pen.endPath()


class CompositeGlyph:
    def draw(self, pen):
        pen.addComponent("simple_glyph", (1, 0, 0, 1, -1, 1))

    def drawPoints(self, pen):
        self.draw(pen)


class MissingComponent(CompositeGlyph):
    def draw(self, pen):
        pen.addComponent("foobar", (1, 0, 0, 1, 0, 0))


class FlippedComponent(CompositeGlyph):
    def draw(self, pen):
        pen.addComponent("simple_glyph", (1, 0, 0, 1, 0, 0))
        pen.addComponent("composite_glyph", (-1, 0, 0, 1, 0, 0))


GLYPHSET = MappingProxyType(
    {
        "simple_glyph": SimpleGlyph(),
        "composite_glyph": CompositeGlyph(),
        "missing_component": MissingComponent(),
        "flipped_component": FlippedComponent(),
    }
)


def _is_point_pen(pen):
    return hasattr(pen, "addPoint")


def _draw(glyph, pen):
    if _is_point_pen(pen):
        # point pen
        glyph.drawPoints(pen)
    else:
        # segment pen
        glyph.draw(pen)


def test_filter_point_pen_positional_identifier():
    rec = RecordingPointPen()
    pen = FilterPointPen(rec)

    pen.beginPath("glyph1")
    pen.addPoint((0, 0), "line", False, None, "pt1")
    pen.endPath()

    assert rec.value == [
        ("beginPath", (), {"identifier": "glyph1"}),
        ("addPoint", ((0, 0), "line", False, None), {"identifier": "pt1"}),
        ("endPath", (), {}),
    ]


@pytest.fixture(params=[DecomposingFilterPen, DecomposingFilterPointPen])
def FilterPen(request):
    return request.param


def _init_rec_and_filter_pens(FilterPenClass, *args, **kwargs):
    rec = out = RecordingPen()
    if _is_point_pen(FilterPenClass):
        out = PointToSegmentPen(rec)
    fpen = FilterPenClass(out, GLYPHSET, *args, **kwargs)
    return rec, fpen


@pytest.mark.parametrize(
    "glyph_name, expected",
    [
        (
            "simple_glyph",
            [
                ("moveTo", ((0, 0),)),
                ("curveTo", ((1, 1), (2, 2), (3, 3))),
                ("closePath", ()),
            ],
        ),
        (
            "composite_glyph",
            [
                ("moveTo", ((-1, 1),)),
                ("curveTo", ((0, 2), (1, 3), (2, 4))),
                ("closePath", ()),
            ],
        ),
        ("missing_component", MissingComponentError),
        (
            "flipped_component",
            [
                ("moveTo", ((0, 0),)),
                ("curveTo", ((1, 1), (2, 2), (3, 3))),
                ("closePath", ()),
                ("moveTo", ((1, 1),)),
                ("curveTo", ((0, 2), (-1, 3), (-2, 4))),
                ("closePath", ()),
            ],
        ),
    ],
)
def test_decomposing_filter_pen(FilterPen, glyph_name, expected):
    rec, fpen = _init_rec_and_filter_pens(FilterPen)
    glyph = GLYPHSET[glyph_name]
    try:
        _draw(glyph, fpen)
    except Exception as e:
        assert isinstance(e, expected)
    else:
        assert rec.value == expected


def test_decomposing_filter_pen_skip_missing(FilterPen, caplog):
    rec, fpen = _init_rec_and_filter_pens(FilterPen, skipMissingComponents=True)
    glyph = GLYPHSET["missing_component"]
    with caplog.at_level(logging.WARNING, logger="fontTools.pens.filterPen"):
        _draw(glyph, fpen)
    assert rec.value == []
    assert "glyph 'foobar' is missing from glyphSet; skipped" in caplog.text


def test_decomposing_filter_pen_reverse_flipped(FilterPen):
    rec, fpen = _init_rec_and_filter_pens(FilterPen, reverseFlipped=True)
    glyph = GLYPHSET["flipped_component"]
    _draw(glyph, fpen)
    assert rec.value == [
        ("moveTo", ((0, 0),)),
        ("curveTo", ((1, 1), (2, 2), (3, 3))),
        ("closePath", ()),
        ("moveTo", ((1, 1),)),
        ("lineTo", ((-2, 4),)),
        ("curveTo", ((-1, 3), (0, 2), (1, 1))),
        ("closePath", ()),
    ]


@pytest.mark.parametrize(
    "decomposeNested, expected",
    [
        (
            True,
            [
                ("addComponent", ("simple_glyph", (1, 0, 0, 1, 0, 0))),
                ("moveTo", ((1, 1),)),
                ("curveTo", ((0, 2), (-1, 3), (-2, 4))),
                ("closePath", ()),
            ],
        ),
        (
            False,
            [
                ("addComponent", ("simple_glyph", (1, 0, 0, 1, 0, 0))),
                ("addComponent", ("simple_glyph", (-1, 0, 0, 1, 1, 1))),
            ],
        ),
    ],
)
def test_decomposing_filter_pen_include_decomposeNested(
    FilterPen, decomposeNested, expected
):
    rec, fpen = _init_rec_and_filter_pens(
        FilterPen, include={"composite_glyph"}, decomposeNested=decomposeNested
    )
    glyph = GLYPHSET["flipped_component"]
    _draw(glyph, fpen)
    assert rec.value == expected


class TestOnCurveFirstPointPen:
    def test_closed_contour_starting_with_offcurve(self):
        """Test that a closed contour starting with an off-curve point gets rotated."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        # Draw a "closed" (not starting with 'move') contour, starting with off-curve
        pen.beginPath()
        pen.addPoint((0, 0), None)
        pen.addPoint((100, 100), "line")  # first on-curve, should become start
        pen.addPoint((200, 0), None)
        pen.addPoint((300, 100), "curve")
        pen.endPath()

        assert rec.value == [
            ("beginPath", (), {}),
            ("addPoint", ((100, 100), "line", False, None), {}),
            ("addPoint", ((200, 0), None, False, None), {}),
            ("addPoint", ((300, 100), "curve", False, None), {}),
            ("addPoint", ((0, 0), None, False, None), {}),
            ("endPath", (), {}),
        ]

    def test_closed_contour_already_starting_with_oncurve(self):
        """Test that a closed contour already starting with on-curve passes through
        unchanged."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        pen.beginPath()
        pen.addPoint((100, 100), "line")
        pen.addPoint((200, 0), None)
        pen.addPoint((300, 100), "curve")
        pen.endPath()

        assert rec.value == [
            ("beginPath", (), {}),
            ("addPoint", ((100, 100), "line", False, None), {}),
            ("addPoint", ((200, 0), None, False, None), {}),
            ("addPoint", ((300, 100), "curve", False, None), {}),
            ("endPath", (), {}),
        ]

    def test_open_contour(self):
        """Test that open contours pass through unchanged."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        # "Open" contour starts with "move"
        pen.beginPath()
        pen.addPoint((0, 0), "move")
        pen.addPoint((100, 100), "line")
        pen.addPoint((200, 200), "curve")
        pen.endPath()

        assert rec.value == [
            ("beginPath", (), {}),
            ("addPoint", ((0, 0), "move", False, None), {}),
            ("addPoint", ((100, 100), "line", False, None), {}),
            ("addPoint", ((200, 200), "curve", False, None), {}),
            ("endPath", (), {}),
        ]

    def test_all_offcurve_points(self):
        """Test contour with all off-curve points (TrueType special case)
        passes through unchanged."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        pen.beginPath()
        pen.addPoint((0, 0), None)
        pen.addPoint((100, 100), None)
        pen.addPoint((200, 0), None)
        pen.endPath()

        # there's no on-curve to rotate to
        assert rec.value == [
            ("beginPath", (), {}),
            ("addPoint", ((0, 0), None, False, None), {}),
            ("addPoint", ((100, 100), None, False, None), {}),
            ("addPoint", ((200, 0), None, False, None), {}),
            ("endPath", (), {}),
        ]

    def test_components_pass_through(self):
        """Test that components pass through unchanged."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        pen.addComponent("a", (1, 0, 0, 1, 10, 20))

        assert rec.value == [
            ("addComponent", ("a", (1, 0, 0, 1, 10, 20)), {}),
        ]

    def test_positional_identifier(self):
        """Test that positional identifier arguments are forwarded."""
        rec = RecordingPointPen()
        pen = OnCurveFirstPointPen(rec)

        pen.beginPath("path1")
        pen.addPoint((0, 0), "move", False, None, "start")
        pen.addPoint((100, 100), "line", False, None, "line-id")
        pen.endPath()

        assert rec.value == [
            ("beginPath", (), {"identifier": "path1"}),
            ("addPoint", ((0, 0), "move", False, None), {"identifier": "start"}),
            ("addPoint", ((100, 100), "line", False, None), {"identifier": "line-id"}),
            ("endPath", (), {}),
        ]


class SimpleGlyphStartingWithOffCurve:
    def drawPoints(self, pen):
        pen.beginPath()
        pen.addPoint((0, 0), None)
        pen.addPoint((50, 50), None)
        pen.addPoint((100, 0), "qcurve")
        pen.addPoint((50, -50), None)
        pen.endPath()


@pytest.mark.parametrize(
    "reverseFlipped", [True, ReverseFlipped.KEEP_START, "keep_start"]
)
def test_decomposing_filter_point_pen_reverse_flipped_keep_start(reverseFlipped):
    """Test DecomposingFilterPointPen with reverseFlipped='keep_start' enum value"""
    from fontTools.pens.filterPen import DecomposingFilterPointPen
    from fontTools.pens.recordingPen import RecordingPointPen

    glyphSet = {"base": SimpleGlyphStartingWithOffCurve()}

    rec = RecordingPointPen()
    fpen = DecomposingFilterPointPen(rec, glyphSet, reverseFlipped=reverseFlipped)
    # Add a flipped component (negative determinant)
    fpen.addComponent("base", (-1, 0, 0, 1, 200, 0))

    # The contour should be reversed but still start with off-curve (segmentType=None)
    assert len(rec.value) == 6  # beginPath + 4 points + endPath
    assert rec.value[0][0] == "beginPath"
    first_point = rec.value[1]
    assert first_point[0] == "addPoint"
    assert first_point[1][:2] == ((200, 0), None)


@pytest.mark.parametrize(
    "reverseFlipped", [ReverseFlipped.ON_CURVE_FIRST, "on_curve_first"]
)
def test_decomposing_filter_point_pen_reverse_flipped_on_curve_first(reverseFlipped):
    """Test DecomposingFilterPointPen with reverseFlipped='on_curve_first' enum value"""
    from fontTools.pens.filterPen import DecomposingFilterPointPen
    from fontTools.pens.recordingPen import RecordingPointPen

    glyphSet = {"base": SimpleGlyphStartingWithOffCurve()}

    rec = RecordingPointPen()
    fpen = DecomposingFilterPointPen(rec, glyphSet, reverseFlipped=reverseFlipped)
    # Add a flipped component (negative determinant)
    fpen.addComponent("base", (-1, 0, 0, 1, 200, 0))

    # The contour should be reversed AND rotated to start with on-curve
    assert len(rec.value) == 6  # beginPath + 4 points + endPath
    assert rec.value[0][0] == "beginPath"
    # First point should now be on-curve
    first_point = rec.value[1]
    assert first_point[0] == "addPoint"
    assert first_point[1][:2] == ((100, 0), "qcurve")


@pytest.mark.parametrize("reverseFlipped", [False, ReverseFlipped.NO, "no"])
def test_decomposing_filter_point_pen_reverse_flipped_no(reverseFlipped):
    """Test DecomposingFilterPointPen with reverseFlipped='no' enum value"""
    from fontTools.pens.filterPen import DecomposingFilterPointPen
    from fontTools.pens.recordingPen import RecordingPointPen

    glyphSet = {"base": SimpleGlyphStartingWithOffCurve()}

    rec = RecordingPointPen()
    fpen = DecomposingFilterPointPen(rec, glyphSet, reverseFlipped=reverseFlipped)
    # Add a flipped component (negative determinant)
    fpen.addComponent("base", (-1, 0, 0, 1, 200, 0))

    # Should not be reversed, so should start with first off-curve but NOT reversed
    assert len(rec.value) == 6
    assert rec.value[0][0] == "beginPath"
    # Check that it's NOT reversed - the points should be transformed but in original order
    first_point = rec.value[1]
    assert first_point[0] == "addPoint"
    assert first_point[1][:2] == ((200, 0), None)



from fontTools.misc.loggingTools import CapturingLogHandler
from fontTools.misc.testTools import parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.ttLib.tables.TupleVariation import (
    log,
    TupleVariation,
    compileSharedTuples,
    decompileSharedTuples,
    compileTupleVariationStore,
    decompileTupleVariationStore,
    inferRegion_,
)
from io import BytesIO
import random
import unittest


def hexencode(s):
    h = hexStr(s).upper()
    return " ".join([h[i : i + 2] for i in range(0, len(h), 2)])


AXES = {
    "wdth": (0.25, 0.375, 0.5),
    "wght": (0.0, 1.0, 1.0),
    "opsz": (-0.75, -0.75, 0.0),
}


# Shared tuples in the 'gvar' table of the Skia font, as printed
# in Apple's TrueType specification.
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6gvar.html
SKIA_GVAR_SHARED_TUPLES_DATA = deHexStr(
    "40 00 00 00 C0 00 00 00 00 00 40 00 00 00 C0 00 "
    "C0 00 C0 00 40 00 C0 00 40 00 40 00 C0 00 40 00"
)

SKIA_GVAR_SHARED_TUPLES = [
    {"wght": 1.0, "wdth": 0.0},
    {"wght": -1.0, "wdth": 0.0},
    {"wght": 0.0, "wdth": 1.0},
    {"wght": 0.0, "wdth": -1.0},
    {"wght": -1.0, "wdth": -1.0},
    {"wght": 1.0, "wdth": -1.0},
    {"wght": 1.0, "wdth": 1.0},
    {"wght": -1.0, "wdth": 1.0},
]


# Tuple Variation Store of uppercase I in the Skia font, as printed in Apple's
# TrueType spec. The actual Skia font uses a different table for uppercase I
# than what is printed in Apple's spec, but we still want to make sure that
# we can parse the data as it appears in the specification.
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6gvar.html
SKIA_GVAR_I_DATA = deHexStr(
    "00 08 00 24 00 33 20 00 00 15 20 01 00 1B 20 02 "
    "00 24 20 03 00 15 20 04 00 26 20 07 00 0D 20 06 "
    "00 1A 20 05 00 40 01 01 01 81 80 43 FF 7E FF 7E "
    "FF 7E FF 7E 00 81 45 01 01 01 03 01 04 01 04 01 "
    "04 01 02 80 40 00 82 81 81 04 3A 5A 3E 43 20 81 "
    "04 0E 40 15 45 7C 83 00 0D 9E F3 F2 F0 F0 F0 F0 "
    "F3 9E A0 A1 A1 A1 9F 80 00 91 81 91 00 0D 0A 0A "
    "09 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0B 80 00 15 81 "
    "81 00 C4 89 00 C4 83 00 0D 80 99 98 96 96 96 96 "
    "99 80 82 83 83 83 81 80 40 FF 18 81 81 04 E6 F9 "
    "10 21 02 81 04 E8 E5 EB 4D DA 83 00 0D CE D3 D4 "
    "D3 D3 D3 D5 D2 CE CC CD CD CD CD 80 00 A1 81 91 "
    "00 0D 07 03 04 02 02 02 03 03 07 07 08 08 08 07 "
    "80 00 09 81 81 00 28 40 00 A4 02 24 24 66 81 04 "
    "08 FA FA FA 28 83 00 82 02 FF FF FF 83 02 01 01 "
    "01 84 91 00 80 06 07 08 08 08 08 0A 07 80 03 FE "
    "FF FF FF 81 00 08 81 82 02 EE EE EE 8B 6D 00"
)


class TupleVariationTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_equal(self):
        var1 = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0), (9, 8), (7, 6)])
        var2 = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0), (9, 8), (7, 6)])
        self.assertEqual(var1, var2)

    def test_equal_differentAxes(self):
        var1 = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0), (9, 8), (7, 6)])
        var2 = TupleVariation({"wght": (0.7, 0.8, 0.9)}, [(0, 0), (9, 8), (7, 6)])
        self.assertNotEqual(var1, var2)

    def test_equal_differentCoordinates(self):
        var1 = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0), (9, 8), (7, 6)])
        var2 = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0), (9, 8)])
        self.assertNotEqual(var1, var2)

    def test_hasImpact_someDeltasNotZero(self):
        axes = {"wght": (0.0, 1.0, 1.0)}
        var = TupleVariation(axes, [(0, 0), (9, 8), (7, 6)])
        self.assertTrue(var.hasImpact())

    def test_hasImpact_allDeltasZero(self):
        axes = {"wght": (0.0, 1.0, 1.0)}
        var = TupleVariation(axes, [(0, 0), (0, 0), (0, 0)])
        self.assertTrue(var.hasImpact())

    def test_hasImpact_allDeltasNone(self):
        axes = {"wght": (0.0, 1.0, 1.0)}
        var = TupleVariation(axes, [None, None, None])
        self.assertFalse(var.hasImpact())

    def test_toXML_badDeltaFormat(self):
        writer = XMLWriter(BytesIO())
        g = TupleVariation(AXES, ["String"])
        with CapturingLogHandler(log, "ERROR") as captor:
            g.toXML(writer, ["wdth"])
        self.assertIn("bad delta format", [r.msg for r in captor.records])
        self.assertEqual(
            [
                "<tuple>",
                '<coord axis="wdth" min="0.25" value="0.375" max="0.5"/>',
                "<!-- bad delta #0 -->",
                "</tuple>",
            ],
            TupleVariationTest.xml_lines(writer),
        )

    def test_toXML_constants(self):
        writer = XMLWriter(BytesIO())
        g = TupleVariation(AXES, [42, None, 23, 0, -17, None])
        g.toXML(writer, ["wdth", "wght", "opsz"])
        self.assertEqual(
            [
                "<tuple>",
                '<coord axis="wdth" min="0.25" value="0.375" max="0.5"/>',
                '<coord axis="wght" value="1.0"/>',
                '<coord axis="opsz" value="-0.75"/>',
                '<delta cvt="0" value="42"/>',
                '<delta cvt="2" value="23"/>',
                '<delta cvt="3" value="0"/>',
                '<delta cvt="4" value="-17"/>',
                "</tuple>",
            ],
            TupleVariationTest.xml_lines(writer),
        )

    def test_toXML_points(self):
        writer = XMLWriter(BytesIO())
        g = TupleVariation(AXES, [(9, 8), None, (7, 6), (0, 0), (-1, -2), None])
        g.toXML(writer, ["wdth", "wght", "opsz"])
        self.assertEqual(
            [
                "<tuple>",
                '<coord axis="wdth" min="0.25" value="0.375" max="0.5"/>',
                '<coord axis="wght" value="1.0"/>',
                '<coord axis="opsz" value="-0.75"/>',
                '<delta pt="0" x="9" y="8"/>',
                '<delta pt="2" x="7" y="6"/>',
                '<delta pt="3" x="0" y="0"/>',
                '<delta pt="4" x="-1" y="-2"/>',
                "</tuple>",
            ],
            TupleVariationTest.xml_lines(writer),
        )

    def test_toXML_allDeltasNone(self):
        writer = XMLWriter(BytesIO())
        axes = {"wght": (0.0, 1.0, 1.0)}
        g = TupleVariation(axes, [None] * 5)
        g.toXML(writer, ["wght", "wdth"])
        self.assertEqual(
            [
                "<tuple>",
                '<coord axis="wght" value="1.0"/>',
                "<!-- no deltas -->",
                "</tuple>",
            ],
            TupleVariationTest.xml_lines(writer),
        )

    def test_toXML_axes_floats(self):
        writer = XMLWriter(BytesIO())
        axes = {
            "wght": (0.0, 0.2999878, 0.7000122),
            "wdth": (0.0, 0.4000244, 0.4000244),
        }
        g = TupleVariation(axes, [None] * 5)
        g.toXML(writer, ["wght", "wdth"])
        self.assertEqual(
            [
                '<coord axis="wght" min="0.0" value="0.3" max="0.7"/>',
                '<coord axis="wdth" value="0.4"/>',
            ],
            TupleVariationTest.xml_lines(writer)[1:3],
        )

    def test_fromXML_badDeltaFormat(self):
        g = TupleVariation({}, [])
        with CapturingLogHandler(log, "WARNING") as captor:
            for name, attrs, content in parseXML('<delta a="1" b="2"/>'):
                g.fromXML(name, attrs, content)
        self.assertIn("bad delta format: a, b", [r.msg for r in captor.records])

    def test_fromXML_constants(self):
        g = TupleVariation({}, [None] * 4)
        for name, attrs, content in parseXML(
            '<coord axis="wdth" min="0.25" value="0.375" max="0.5"/>'
            '<coord axis="wght" value="1.0"/>'
            '<coord axis="opsz" value="-0.75"/>'
            '<delta cvt="1" value="42"/>'
            '<delta cvt="2" value="-23"/>'
        ):
            g.fromXML(name, attrs, content)
        self.assertEqual(AXES, g.axes)
        self.assertEqual([None, 42, -23, None], g.coordinates)

    def test_fromXML_points(self):
        g = TupleVariation({}, [None] * 4)
        for name, attrs, content in parseXML(
            '<coord axis="wdth" min="0.25" value="0.375" max="0.5"/>'
            '<coord axis="wght" value="1.0"/>'
            '<coord axis="opsz" value="-0.75"/>'
            '<delta pt="1" x="33" y="44"/>'
            '<delta pt="2" x="-2" y="170"/>'
        ):
            g.fromXML(name, attrs, content)
        self.assertEqual(AXES, g.axes)
        self.assertEqual([None, (33, 44), (-2, 170), None], g.coordinates)

    def test_fromXML_axes_floats(self):
        g = TupleVariation({}, [None] * 4)
        for name, attrs, content in parseXML(
            '<coord axis="wght" min="0.0" value="0.3" max="0.7"/>'
            '<coord axis="wdth" value="0.4"/>'
        ):
            g.fromXML(name, attrs, content)

        self.assertEqual(g.axes["wght"][0], 0)
        self.assertAlmostEqual(g.axes["wght"][1], 0.2999878)
        self.assertAlmostEqual(g.axes["wght"][2], 0.7000122)

        self.assertEqual(g.axes["wdth"][0], 0)
        self.assertAlmostEqual(g.axes["wdth"][1], 0.4000244)
        self.assertAlmostEqual(g.axes["wdth"][2], 0.4000244)

    def test_compile_sharedPeaks_nonIntermediate_sharedPoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [(7, 4), (8, 5), (9, 6)]
        )
        axisTags = ["wght", "wdth"]
        sharedPeakIndices = {var.compileCoord(axisTags): 0x77}
        tup, deltas = var.compile(axisTags, sharedPeakIndices, pointData=b"")
        # len(deltas)=8; flags=None; tupleIndex=0x77
        # embeddedPeaks=[]; intermediateCoord=[]
        self.assertEqual("00 08 00 77", hexencode(tup))
        self.assertEqual(
            "02 07 08 09 " "02 04 05 06",  # deltaX: [7, 8, 9]  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_sharedPeaks_intermediate_sharedPoints(self):
        var = TupleVariation(
            {"wght": (0.3, 0.5, 0.7), "wdth": (0.1, 0.8, 0.9)}, [(7, 4), (8, 5), (9, 6)]
        )
        axisTags = ["wght", "wdth"]
        sharedPeakIndices = {var.compileCoord(axisTags): 0x77}
        tup, deltas = var.compile(axisTags, sharedPeakIndices, pointData=b"")
        # len(deltas)=8; flags=INTERMEDIATE_REGION; tupleIndex=0x77
        # embeddedPeak=[]; intermediateCoord=[(0.3, 0.1), (0.7, 0.9)]
        self.assertEqual("00 08 40 77 13 33 06 66 2C CD 39 9A", hexencode(tup))
        self.assertEqual(
            "02 07 08 09 " "02 04 05 06",  # deltaX: [7, 8, 9]  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_sharedPeaks_nonIntermediate_privatePoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [(7, 4), (8, 5), (9, 6)]
        )
        axisTags = ["wght", "wdth"]
        sharedPeakIndices = {var.compileCoord(axisTags): 0x77}
        tup, deltas = var.compile(axisTags, sharedPeakIndices)
        # len(deltas)=9; flags=PRIVATE_POINT_NUMBERS; tupleIndex=0x77
        # embeddedPeak=[]; intermediateCoord=[]
        self.assertEqual("00 09 20 77", hexencode(tup))
        self.assertEqual(
            "00 "  # all points in glyph
            "02 07 08 09 "  # deltaX: [7, 8, 9]
            "02 04 05 06",  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_sharedPeaks_intermediate_privatePoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 1.0), "wdth": (0.0, 0.8, 1.0)}, [(7, 4), (8, 5), (9, 6)]
        )
        axisTags = ["wght", "wdth"]
        sharedPeakIndices = {var.compileCoord(axisTags): 0x77}
        tuple, deltas = var.compile(axisTags, sharedPeakIndices)
        # len(deltas)=9; flags=PRIVATE_POINT_NUMBERS; tupleIndex=0x77
        # embeddedPeak=[]; intermediateCoord=[(0.0, 0.0), (1.0, 1.0)]
        self.assertEqual("00 09 60 77 00 00 00 00 40 00 40 00", hexencode(tuple))
        self.assertEqual(
            "00 "  # all points in glyph
            "02 07 08 09 "  # deltaX: [7, 8, 9]
            "02 04 05 06",  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_nonIntermediate_sharedPoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [(7, 4), (8, 5), (9, 6)]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"], pointData=b"")
        # len(deltas)=8; flags=EMBEDDED_PEAK_TUPLE
        # embeddedPeak=[(0.5, 0.8)]; intermediateCoord=[]
        self.assertEqual("00 08 80 00 20 00 33 33", hexencode(tup))
        self.assertEqual(
            "02 07 08 09 " "02 04 05 06",  # deltaX: [7, 8, 9]  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_nonIntermediate_sharedConstants(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [3, 1, 4]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"], pointData=b"")
        # len(deltas)=4; flags=EMBEDDED_PEAK_TUPLE
        # embeddedPeak=[(0.5, 0.8)]; intermediateCoord=[]
        self.assertEqual("00 04 80 00 20 00 33 33", hexencode(tup))
        self.assertEqual("02 03 01 04", hexencode(deltas))  # delta: [3, 1, 4]

    def test_compile_embeddedPeak_intermediate_sharedPoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 1.0), "wdth": (0.0, 0.8, 0.8)}, [(7, 4), (8, 5), (9, 6)]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"], pointData=b"")
        # len(deltas)=8; flags=EMBEDDED_PEAK_TUPLE
        # embeddedPeak=[(0.5, 0.8)]; intermediateCoord=[(0.0, 0.0), (1.0, 0.8)]
        self.assertEqual(
            "00 08 C0 00 20 00 33 33 00 00 00 00 40 00 33 33", hexencode(tup)
        )
        self.assertEqual(
            "02 07 08 09 " "02 04 05 06",  # deltaX: [7, 8, 9]  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_nonIntermediate_privatePoints(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [(7, 4), (8, 5), (9, 6)]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"])
        # len(deltas)=9; flags=PRIVATE_POINT_NUMBERS|EMBEDDED_PEAK_TUPLE
        # embeddedPeak=[(0.5, 0.8)]; intermediateCoord=[]
        self.assertEqual("00 09 A0 00 20 00 33 33", hexencode(tup))
        self.assertEqual(
            "00 "  # all points in glyph
            "02 07 08 09 "  # deltaX: [7, 8, 9]
            "02 04 05 06",  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_nonIntermediate_privateConstants(self):
        var = TupleVariation(
            {"wght": (0.0, 0.5, 0.5), "wdth": (0.0, 0.8, 0.8)}, [7, 8, 9]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"])
        # len(deltas)=5; flags=PRIVATE_POINT_NUMBERS|EMBEDDED_PEAK_TUPLE
        # embeddedPeak=[(0.5, 0.8)]; intermediateCoord=[]
        self.assertEqual("00 05 A0 00 20 00 33 33", hexencode(tup))
        self.assertEqual(
            "00 " "02 07 08 09",  # all points in glyph  # delta: [7, 8, 9]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_intermediate_privatePoints(self):
        var = TupleVariation(
            {"wght": (0.4, 0.5, 0.6), "wdth": (0.7, 0.8, 0.9)}, [(7, 4), (8, 5), (9, 6)]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"])
        # len(deltas)=9;
        # flags=PRIVATE_POINT_NUMBERS|INTERMEDIATE_REGION|EMBEDDED_PEAK_TUPLE
        # embeddedPeak=(0.5, 0.8); intermediateCoord=[(0.4, 0.7), (0.6, 0.9)]
        self.assertEqual(
            "00 09 E0 00 20 00 33 33 19 9A 2C CD 26 66 39 9A", hexencode(tup)
        )
        self.assertEqual(
            "00 "  # all points in glyph
            "02 07 08 09 "  # deltaX: [7, 8, 9]
            "02 04 05 06",  # deltaY: [4, 5, 6]
            hexencode(deltas),
        )

    def test_compile_embeddedPeak_intermediate_privateConstants(self):
        var = TupleVariation(
            {"wght": (0.4, 0.5, 0.6), "wdth": (0.7, 0.8, 0.9)}, [7, 8, 9]
        )
        tup, deltas = var.compile(axisTags=["wght", "wdth"])
        # len(deltas)=5;
        # flags=PRIVATE_POINT_NUMBERS|INTERMEDIATE_REGION|EMBEDDED_PEAK_TUPLE
        # embeddedPeak=(0.5, 0.8); intermediateCoord=[(0.4, 0.7), (0.6, 0.9)]
        self.assertEqual(
            "00 05 E0 00 20 00 33 33 19 9A 2C CD 26 66 39 9A", hexencode(tup)
        )
        self.assertEqual(
            "00 " "02 07 08 09",  # all points in glyph  # delta: [7, 8, 9]
            hexencode(deltas),
        )

    def test_compileCoord(self):
        var = TupleVariation(
            {"wght": (-1.0, -1.0, -1.0), "wdth": (0.4, 0.5, 0.6)}, [None] * 4
        )
        self.assertEqual("C0 00 20 00", hexencode(var.compileCoord(["wght", "wdth"])))
        self.assertEqual("20 00 C0 00", hexencode(var.compileCoord(["wdth", "wght"])))
        self.assertEqual("C0 00", hexencode(var.compileCoord(["wght"])))

    def test_compileIntermediateCoord(self):
        var = TupleVariation(
            {"wght": (-1.0, -1.0, 0.0), "wdth": (0.4, 0.5, 0.6)}, [None] * 4
        )
        self.assertEqual(
            "C0 00 19 9A 00 00 26 66",
            hexencode(var.compileIntermediateCoord(["wght", "wdth"])),
        )
        self.assertEqual(
            "19 9A C0 00 26 66 00 00",
            hexencode(var.compileIntermediateCoord(["wdth", "wght"])),
        )
        self.assertEqual(None, var.compileIntermediateCoord(["wght"]))
        self.assertEqual(
            "19 9A 26 66", hexencode(var.compileIntermediateCoord(["wdth"]))
        )

    def test_decompileCoord(self):
        decompileCoord = TupleVariation.decompileCoord_
        data = deHexStr("DE AD C0 00 20 00 DE AD")
        self.assertEqual(
            ({"wght": -1.0, "wdth": 0.5}, 6), decompileCoord(["wght", "wdth"], data, 2)
        )

    def test_decompileCoord_roundTrip(self):
        # Make sure we are not affected by https://github.com/fonttools/fonttools/issues/286
        data = deHexStr("7F B9 80 35")
        values, _ = TupleVariation.decompileCoord_(["wght", "wdth"], data, 0)
        axisValues = {axis: (val, val, val) for axis, val in values.items()}
        var = TupleVariation(axisValues, [None] * 4)
        self.assertEqual("7F B9 80 35", hexencode(var.compileCoord(["wght", "wdth"])))

    def test_compilePoints(self):
        compilePoints = lambda p: TupleVariation.compilePoints(set(p))
        self.assertEqual("00", hexencode(compilePoints(set())))  # all points in glyph
        self.assertEqual("01 00 07", hexencode(compilePoints([7])))
        self.assertEqual("01 80 FF FF", hexencode(compilePoints([65535])))
        self.assertEqual("02 01 09 06", hexencode(compilePoints([9, 15])))
        self.assertEqual(
            "06 05 07 01 F7 02 01 F2",
            hexencode(compilePoints([7, 8, 255, 257, 258, 500])),
        )
        self.assertEqual("03 01 07 01 80 01 EC", hexencode(compilePoints([7, 8, 500])))
        self.assertEqual(
            "04 01 07 01 81 BE E7 0C 0F",
            hexencode(compilePoints([7, 8, 0xBEEF, 0xCAFE])),
        )
        self.maxDiff = None
        self.assertEqual(
            "81 2C"
            + " 7F 00"  # 300 points (0x12c) in total
            + (127 * " 01")
            + " 7F"  # first run, contains 128 points: [0 .. 127]
            + (128 * " 01")
            + " 2B"  # second run, contains 128 points: [128 .. 255]
            + (44 * " 01"),  # third run, contains 44 points: [256 .. 299]
            hexencode(compilePoints(range(300))),
        )
        self.assertEqual(
            "81 8F"
            + " 7F 00"  # 399 points (0x18f) in total
            + (127 * " 01")
            + " 7F"  # first run, contains 128 points: [0 .. 127]
            + (128 * " 01")
            + " 7F"  # second run, contains 128 points: [128 .. 255]
            + (128 * " 01")
            + " 0E"  # third run, contains 128 points: [256 .. 383]
            + (15 * " 01"),  # fourth run, contains 15 points: [384 .. 398]
            hexencode(compilePoints(range(399))),
        )

    def test_decompilePoints(self):
        numPointsInGlyph = 65536
        allPoints = list(range(numPointsInGlyph))

        def decompilePoints(data, offset):
            points, offset = TupleVariation.decompilePoints_(
                numPointsInGlyph, deHexStr(data), offset, "gvar"
            )
            # Conversion to list needed for Python 3.
            return (list(points), offset)

        # all points in glyph
        self.assertEqual((allPoints, 1), decompilePoints("00", 0))
        # all points in glyph (in overly verbose encoding, not explicitly prohibited by spec)
        self.assertEqual((allPoints, 2), decompilePoints("80 00", 0))
        # 2 points; first run: [9, 9+6]
        self.assertEqual(([9, 15], 4), decompilePoints("02 01 09 06", 0))
        # 2 points; first run: [0xBEEF, 0xCAFE]. (0x0C0F = 0xCAFE - 0xBEEF)
        self.assertEqual(([0xBEEF, 0xCAFE], 6), decompilePoints("02 81 BE EF 0C 0F", 0))
        # 1 point; first run: [7]
        self.assertEqual(([7], 3), decompilePoints("01 00 07", 0))
        # 1 point; first run: [7] in overly verbose encoding
        self.assertEqual(([7], 4), decompilePoints("01 80 00 07", 0))
        # 1 point; first run: [65535]; requires words to be treated as unsigned numbers
        self.assertEqual(([65535], 4), decompilePoints("01 80 FF FF", 0))
        # 4 points; first run: [7, 8]; second run: [255, 257]. 257 is stored in delta-encoded bytes (0xFF + 2).
        self.assertEqual(
            ([7, 8, 263, 265], 7), decompilePoints("04 01 07 01 01 FF 02", 0)
        )
        # combination of all encodings, preceded and followed by 4 bytes of unused data
        data = "DE AD DE AD 04 01 07 01 81 BE E7 0C 0F DE AD DE AD"
        self.assertEqual(([7, 8, 0xBEEF, 0xCAFE], 13), decompilePoints(data, 4))
        self.assertSetEqual(
            set(range(300)),
            set(
                decompilePoints(
                    "81 2C"
                    + " 7F 00"  # 300 points (0x12c) in total
                    + (127 * " 01")
                    + " 7F"  # first run, contains 128 points: [0 .. 127]
                    + (128 * " 01")
                    + " AB"  # second run, contains 128 points: [128 .. 255]
                    + (44 * " 00 01"),  # third run, contains 44 points: [256 .. 299]
                    0,
                )[0]
            ),
        )
        self.assertSetEqual(
            set(range(399)),
            set(
                decompilePoints(
                    "81 8F"
                    + " 7F 00"  # 399 points (0x18f) in total
                    + (127 * " 01")
                    + " 7F"  # first run, contains 128 points: [0 .. 127]
                    + (128 * " 01")
                    + " FF"  # second run, contains 128 points: [128 .. 255]
                    + (128 * " 00 01")
                    + " 8E"  # third run, contains 128 points: [256 .. 383]
                    + (15 * " 00 01"),  # fourth run, contains 15 points: [384 .. 398]
                    0,
                )[0]
            ),
        )

    def test_decompilePoints_shouldAcceptBadPointNumbers(self):
        decompilePoints = TupleVariation.decompilePoints_
        # 2 points; first run: [3, 9].
        numPointsInGlyph = 8
        with CapturingLogHandler(log, "WARNING") as captor:
            decompilePoints(numPointsInGlyph, deHexStr("02 01 03 06"), 0, "cvar")
        self.assertIn(
            "point 9 out of range in 'cvar' table", [r.msg for r in captor.records]
        )

    def test_decompilePoints_roundTrip(self):
        numPointsInGlyph = (
            500  # greater than 255, so we also exercise code path for 16-bit encoding
        )
        compile = lambda points: TupleVariation.compilePoints(points)
        decompile = lambda data: set(
            TupleVariation.decompilePoints_(numPointsInGlyph, data, 0, "gvar")[0]
        )
        for i in range(50):
            points = set(random.sample(range(numPointsInGlyph), 30))
            self.assertSetEqual(
                points,
                decompile(compile(points)),
                "failed round-trip decompile/compilePoints; points=%s" % points,
            )
        allPoints = set(range(numPointsInGlyph))
        self.assertSetEqual(allPoints, decompile(compile(allPoints)))
        self.assertSetEqual(allPoints, decompile(compile(set())))

    def test_compileDeltas_points(self):
        var = TupleVariation({}, [None, (1, 0), (2, 0), None, (4, 0), None])
        # deltaX for points: [1, 2, 4]; deltaY for points: [0, 0, 0]
        self.assertEqual("02 01 02 04 82", hexencode(var.compileDeltas()))

    def test_compileDeltas_constants(self):
        var = TupleVariation({}, [None, 1, 2, None, 4, None])
        # delta for cvts: [1, 2, 4]
        self.assertEqual("02 01 02 04", hexencode(var.compileDeltas()))

    def test_compileDeltaValues(self):
        compileDeltaValues = lambda values, optimizeSize=True: hexencode(
            TupleVariation.compileDeltaValues_(values, optimizeSize=optimizeSize)
        )
        # zeroes
        self.assertEqual("80", compileDeltaValues([0]))
        self.assertEqual("BF", compileDeltaValues([0] * 64))
        self.assertEqual("BF 80", compileDeltaValues([0] * 65))
        self.assertEqual("BF A3", compileDeltaValues([0] * 100))
        self.assertEqual("BF BF BF BF", compileDeltaValues([0] * 256))
        # bytes
        self.assertEqual("00 01", compileDeltaValues([1]))
        self.assertEqual(
            "06 01 02 03 7F 80 FF FE", compileDeltaValues([1, 2, 3, 127, -128, -1, -2])
        )
        self.assertEqual("3F" + (64 * " 7F"), compileDeltaValues([127] * 64))
        self.assertEqual("3F" + (64 * " 7F") + " 00 7F", compileDeltaValues([127] * 65))
        # words
        self.assertEqual("40 66 66", compileDeltaValues([0x6666]))
        self.assertEqual(
            "43 66 66 7F FF FF FF 80 00",
            compileDeltaValues([0x6666, 32767, -1, -32768]),
        )
        self.assertEqual("7F" + (64 * " 11 22"), compileDeltaValues([0x1122] * 64))
        self.assertEqual(
            "7F" + (64 * " 11 22") + " 40 11 22", compileDeltaValues([0x1122] * 65)
        )
        # bytes, zeroes, bytes: a single zero is more compact when encoded as part of the bytes run
        self.assertEqual(
            "04 7F 7F 00 7F 7F", compileDeltaValues([127, 127, 0, 127, 127])
        )
        self.assertEqual(
            "01 7F 7F 81 01 7F 7F", compileDeltaValues([127, 127, 0, 0, 127, 127])
        )
        self.assertEqual(
            "01 7F 7F 82 01 7F 7F", compileDeltaValues([127, 127, 0, 0, 0, 127, 127])
        )
        self.assertEqual(
            "01 7F 7F 83 01 7F 7F", compileDeltaValues([127, 127, 0, 0, 0, 0, 127, 127])
        )
        # bytes, zeroes
        self.assertEqual("01 01 00", compileDeltaValues([1, 0]))
        self.assertEqual("01 01 00", compileDeltaValues([1, 0], optimizeSize=False))
        self.assertEqual("00 01 81", compileDeltaValues([1, 0, 0]))
        self.assertEqual(
            "02 01 00 00", compileDeltaValues([1, 0, 0], optimizeSize=False)
        )
        # words, bytes, words: a single byte is more compact when encoded as part of the words run
        self.assertEqual(
            "42 66 66 00 02 77 77", compileDeltaValues([0x6666, 2, 0x7777])
        )
        self.assertEqual(
            "42 66 66 00 02 77 77",
            compileDeltaValues([0x6666, 2, 0x7777], optimizeSize=False),
        )
        self.assertEqual(
            "40 66 66 01 02 02 40 77 77", compileDeltaValues([0x6666, 2, 2, 0x7777])
        )
        self.assertEqual(
            "43 66 66 00 02 00 02 77 77",
            compileDeltaValues([0x6666, 2, 2, 0x7777], optimizeSize=False),
        )
        # words, zeroes, words
        self.assertEqual(
            "40 66 66 80 40 77 77", compileDeltaValues([0x6666, 0, 0x7777])
        )
        self.assertEqual(
            "42 66 66 00 00 77 77",
            compileDeltaValues([0x6666, 0, 0x7777], optimizeSize=False),
        )
        self.assertEqual(
            "40 66 66 81 40 77 77", compileDeltaValues([0x6666, 0, 0, 0x7777])
        )
        self.assertEqual(
            "40 66 66 82 40 77 77", compileDeltaValues([0x6666, 0, 0, 0, 0x7777])
        )
        # words, zeroes, bytes
        self.assertEqual(
            "40 66 66 80 02 01 02 03", compileDeltaValues([0x6666, 0, 1, 2, 3])
        )
        self.assertEqual(
            "40 66 66 81 02 01 02 03", compileDeltaValues([0x6666, 0, 0, 1, 2, 3])
        )
        self.assertEqual(
            "40 66 66 82 02 01 02 03", compileDeltaValues([0x6666, 0, 0, 0, 1, 2, 3])
        )
        # words, zeroes
        self.assertEqual("40 66 66 80", compileDeltaValues([0x6666, 0]))
        self.assertEqual("40 66 66 81", compileDeltaValues([0x6666, 0, 0]))

    def test_decompileDeltas(self):
        decompileDeltas = TupleVariation.decompileDeltas_
        # 83 = zero values (0x80), count = 4 (1 + 0x83 & 0x3F)
        self.assertEqual(([0, 0, 0, 0], 1), decompileDeltas(4, deHexStr("83"), 0))
        # 41 01 02 FF FF = signed 16-bit values (0x40), count = 2 (1 + 0x41 & 0x3F)
        self.assertEqual(
            ([258, -1], 5), decompileDeltas(2, deHexStr("41 01 02 FF FF"), 0)
        )
        # 01 81 07 = signed 8-bit values, count = 2 (1 + 0x01 & 0x3F)
        self.assertEqual(([-127, 7], 3), decompileDeltas(2, deHexStr("01 81 07"), 0))
        # combination of all three encodings, preceded and followed by 4 bytes of unused data
        data = deHexStr("DE AD BE EF 83 40 01 02 01 81 80 DE AD BE EF")
        self.assertEqual(
            ([0, 0, 0, 0, 258, -127, -128], 11), decompileDeltas(7, data, 4)
        )

    def test_decompileDeltas_roundTrip(self):
        numDeltas = 30
        compile = TupleVariation.compileDeltaValues_
        decompile = lambda data: TupleVariation.decompileDeltas_(numDeltas, data, 0)[0]
        for i in range(50):
            deltas = random.sample(range(-128, 127), 10)
            deltas.extend(random.sample(range(-32768, 32767), 10))
            deltas.extend([0] * 10)
            random.shuffle(deltas)
            self.assertListEqual(deltas, decompile(compile(deltas)))

    def test_compileSharedTuples(self):
        # Below, the peak coordinate {"wght": 1.0, "wdth": 0.8} appears
        # three times (most frequent sorted first); {"wght": 1.0, "wdth": 0.5}
        # and {"wght": 1.0, "wdth": 0.7} both appears two times (tie) and
        # are sorted alphanumerically to ensure determinism.
        # The peak coordinate {"wght": 1.0, "wdth": 0.9} appears only once
        # and is thus ignored.
        # Because the start and end of variation ranges is not encoded
        # into the shared pool, they should get ignored.
        deltas = [None] * 4
        variations = [
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.5, 0.7, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.2, 0.7, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.2, 0.8, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.3, 0.5, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.3, 0.8, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.3, 0.9, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.4, 0.8, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (0.5, 0.5, 1.0)}, deltas),
        ]
        result = compileSharedTuples(["wght", "wdth"], variations)
        self.assertEqual(
            [hexencode(c) for c in result],
            ["40 00 33 33", "40 00 20 00", "40 00 2C CD"],
        )

    def test_decompileSharedTuples_Skia(self):
        sharedTuples = decompileSharedTuples(
            axisTags=["wght", "wdth"],
            sharedTupleCount=8,
            data=SKIA_GVAR_SHARED_TUPLES_DATA,
            offset=0,
        )
        self.assertEqual(sharedTuples, SKIA_GVAR_SHARED_TUPLES)

    def test_decompileSharedTuples_empty(self):
        self.assertEqual(decompileSharedTuples(["wght"], 0, b"", 0), [])

    def test_compileTupleVariationStore_allVariationsRedundant(self):
        axes = {"wght": (0.3, 0.4, 0.5), "opsz": (0.7, 0.8, 0.9)}
        variations = [
            TupleVariation(axes, [None] * 4),
            TupleVariation(axes, [None] * 4),
            TupleVariation(axes, [None] * 4),
        ]
        self.assertEqual(
            compileTupleVariationStore(
                variations,
                pointCount=8,
                axisTags=["wght", "opsz"],
                sharedTupleIndices={},
            ),
            (0, b"", b""),
        )

    def test_compileTupleVariationStore_noVariations(self):
        self.assertEqual(
            compileTupleVariationStore(
                variations=[],
                pointCount=8,
                axisTags=["wght", "opsz"],
                sharedTupleIndices={},
            ),
            (0, b"", b""),
        )

    def test_compileTupleVariationStore_roundTrip_cvar(self):
        deltas = [1, 2, 3, 4]
        variations = [
            TupleVariation({"wght": (0.5, 1.0, 1.0), "wdth": (1.0, 1.0, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (1.0, 1.0, 1.0)}, deltas),
        ]
        tupleVariationCount, tuples, data = compileTupleVariationStore(
            variations, pointCount=4, axisTags=["wght", "wdth"], sharedTupleIndices={}
        )
        self.assertEqual(
            decompileTupleVariationStore(
                "cvar",
                ["wght", "wdth"],
                tupleVariationCount,
                pointCount=4,
                sharedTuples={},
                data=(tuples + data),
                pos=0,
                dataPos=len(tuples),
            ),
            variations,
        )

    def test_compileTupleVariationStore_roundTrip_gvar(self):
        deltas = [(1, 1), (2, 2), (3, 3), (4, 4)]
        variations = [
            TupleVariation({"wght": (0.5, 1.0, 1.0), "wdth": (1.0, 1.0, 1.0)}, deltas),
            TupleVariation({"wght": (1.0, 1.0, 1.0), "wdth": (1.0, 1.0, 1.0)}, deltas),
        ]
        tupleVariationCount, tuples, data = compileTupleVariationStore(
            variations, pointCount=4, axisTags=["wght", "wdth"], sharedTupleIndices={}
        )
        self.assertEqual(
            decompileTupleVariationStore(
                "gvar",
                ["wght", "wdth"],
                tupleVariationCount,
                pointCount=4,
                sharedTuples={},
                data=(tuples + data),
                pos=0,
                dataPos=len(tuples),
            ),
            variations,
        )

    def test_decompileTupleVariationStore_Skia_I(self):
        tvar = decompileTupleVariationStore(
            tableTag="gvar",
            axisTags=["wght", "wdth"],
            tupleVariationCount=8,
            pointCount=18,
            sharedTuples=SKIA_GVAR_SHARED_TUPLES,
            data=SKIA_GVAR_I_DATA,
            pos=4,
            dataPos=36,
        )
        self.assertEqual(len(tvar), 8)
        self.assertEqual(tvar[0].axes, {"wght": (0.0, 1.0, 1.0)})
        self.assertEqual(
            " ".join(["%d,%d" % c for c in tvar[0].coordinates]),
            "257,0 -127,0 -128,58 -130,90 -130,62 -130,67 -130,32 -127,0 "
            "257,0 259,14 260,64 260,21 260,69 258,124 0,0 130,0 0,0 0,0",
        )

    def test_decompileTupleVariationStore_empty(self):
        self.assertEqual(
            decompileTupleVariationStore(
                tableTag="gvar",
                axisTags=[],
                tupleVariationCount=0,
                pointCount=5,
                sharedTuples=[],
                data=b"",
                pos=4,
                dataPos=4,
            ),
            [],
        )

    def test_getTupleSize(self):
        getTupleSize = TupleVariation.getTupleSize_
        numAxes = 3
        self.assertEqual(4 + numAxes * 2, getTupleSize(0x8042, numAxes))
        self.assertEqual(4 + numAxes * 4, getTupleSize(0x4077, numAxes))
        self.assertEqual(4, getTupleSize(0x2077, numAxes))
        self.assertEqual(4, getTupleSize(11, numAxes))

    def test_inferRegion(self):
        start, end = inferRegion_({"wght": -0.3, "wdth": 0.7})
        self.assertEqual(start, {"wght": -0.3, "wdth": 0.0})
        self.assertEqual(end, {"wght": 0.0, "wdth": 0.7})

    @staticmethod
    def xml_lines(writer):
        content = writer.file.getvalue().decode("utf-8")
        return [line.strip() for line in content.splitlines()][1:]

    def test_getCoordWidth(self):
        empty = TupleVariation({}, [])
        self.assertEqual(empty.getCoordWidth(), 0)

        empty = TupleVariation({}, [None])
        self.assertEqual(empty.getCoordWidth(), 0)

        gvarTuple = TupleVariation({}, [None, (0, 0)])
        self.assertEqual(gvarTuple.getCoordWidth(), 2)

        cvarTuple = TupleVariation({}, [None, 0])
        self.assertEqual(cvarTuple.getCoordWidth(), 1)

        cvarTuple.coordinates[1] *= 1.0
        self.assertEqual(cvarTuple.getCoordWidth(), 1)

        with self.assertRaises(TypeError):
            TupleVariation({}, [None, "a"]).getCoordWidth()

    def test_scaleDeltas_cvar(self):
        var = TupleVariation({}, [100, None])

        var.scaleDeltas(1.0)
        self.assertEqual(var.coordinates, [100, None])

        var.scaleDeltas(0.333)
        self.assertAlmostEqual(var.coordinates[0], 33.3)
        self.assertIsNone(var.coordinates[1])

        var.scaleDeltas(0.0)
        self.assertEqual(var.coordinates, [0, None])

    def test_scaleDeltas_gvar(self):
        var = TupleVariation({}, [(100, 200), None])

        var.scaleDeltas(1.0)
        self.assertEqual(var.coordinates, [(100, 200), None])

        var.scaleDeltas(0.333)
        self.assertAlmostEqual(var.coordinates[0][0], 33.3)
        self.assertAlmostEqual(var.coordinates[0][1], 66.6)
        self.assertIsNone(var.coordinates[1])

        var.scaleDeltas(0.0)
        self.assertEqual(var.coordinates, [(0, 0), None])

    def test_roundDeltas_cvar(self):
        var = TupleVariation({}, [55.5, None, 99.9])
        var.roundDeltas()
        self.assertEqual(var.coordinates, [56, None, 100])

    def test_roundDeltas_gvar(self):
        var = TupleVariation({}, [(55.5, 100.0), None, (99.9, 100.0)])
        var.roundDeltas()
        self.assertEqual(var.coordinates, [(56, 100), None, (100, 100)])

    def test_calcInferredDeltas(self):
        var = TupleVariation({}, [(0, 0), None, None, None])
        coords = [(1, 1), (1, 1), (1, 1), (1, 1)]

        var.calcInferredDeltas(coords, [])

        self.assertEqual(var.coordinates, [(0, 0), (0, 0), (0, 0), (0, 0)])

    def test_calcInferredDeltas_invalid(self):
        # cvar tuples can't have inferred deltas
        with self.assertRaises(TypeError):
            TupleVariation({}, [0]).calcInferredDeltas([], [])

        # origCoords must have same length as self.coordinates
        with self.assertRaises(ValueError):
            TupleVariation({}, [(0, 0), None]).calcInferredDeltas([], [])

        # at least 4 phantom points required
        with self.assertRaises(AssertionError):
            TupleVariation({}, [(0, 0), None]).calcInferredDeltas([(0, 0), (0, 0)], [])

        with self.assertRaises(AssertionError):
            TupleVariation({}, [(0, 0)] + [None] * 5).calcInferredDeltas(
                [(0, 0)] * 6, [1, 0]  # endPts not in increasing order
            )

    def test_optimize(self):
        var = TupleVariation({"wght": (0.0, 1.0, 1.0)}, [(0, 0)] * 5)

        var.optimize([(0, 0)] * 5, [0])

        self.assertEqual(var.coordinates, [None, None, None, None, None])

    def test_sum_deltas_gvar(self):
        var1 = TupleVariation(
            {},
            [
                (-20, 0),
                (-20, 0),
                (20, 0),
                (20, 0),
                (0, 0),
                (0, 0),
                (0, 0),
                (0, 0),
            ],
        )
        var2 = TupleVariation(
            {},
            [
                (-10, 0),
                (-10, 0),
                (10, 0),
                (10, 0),
                (0, 0),
                (20, 0),
                (0, 0),
                (0, 0),
            ],
        )

        var1 += var2

        self.assertEqual(
            var1.coordinates,
            [
                (-30, 0),
                (-30, 0),
                (30, 0),
                (30, 0),
                (0, 0),
                (20, 0),
                (0, 0),
                (0, 0),
            ],
        )

    def test_sum_deltas_gvar_invalid_length(self):
        var1 = TupleVariation({}, [(1, 2)])
        var2 = TupleVariation({}, [(1, 2), (3, 4)])

        with self.assertRaisesRegex(ValueError, "deltas with different lengths"):
            var1 += var2

    def test_sum_deltas_gvar_with_inferred_points(self):
        var1 = TupleVariation({}, [(1, 2), None])
        var2 = TupleVariation({}, [(2, 3), None])

        with self.assertRaisesRegex(ValueError, "deltas with inferred points"):
            var1 += var2

    def test_sum_deltas_cvar(self):
        axes = {"wght": (0.0, 1.0, 1.0)}
        var1 = TupleVariation(axes, [0, 1, None, None])
        var2 = TupleVariation(axes, [None, 2, None, 3])
        var3 = TupleVariation(axes, [None, None, None, 4])

        var1 += var2
        var1 += var3

        self.assertEqual(var1.coordinates, [0, 3, None, 7])


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from .otBase import BaseTTXConverter


class table_M_A_T_H_(BaseTTXConverter):
    """Mathematical Typesetting table

    The ``MATH`` table contains a variety of information needed to
    typeset glyphs in mathematical formulas and expressions.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/math
    """

    pass



import contextlib
import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple

import pytest
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools.fontBuilder import FontBuilder
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables.otBase import OTTableWriter


def test_main(tmpdir: Path):
    """Check that calling the main function on an input TTF works."""
    glyphs = ".notdef space A Aacute B D".split()
    features = """
    @A = [A Aacute];
    @B = [B D];
    feature kern {
        pos @A @B -50;
    } kern;
    """
    fb = FontBuilder(1000)
    fb.setupGlyphOrder(glyphs)
    addOpenTypeFeaturesFromString(fb.font, features)
    input = tmpdir / "in.ttf"
    fb.save(str(input))
    output = tmpdir / "out.ttf"
    args = [
        "--gpos-compression-level",
        "5",
        str(input),
        "-o",
        str(output),
    ]
    from fontTools.otlLib.optimize import main

    ret = main(args)
    assert ret in (0, None)
    assert output.exists()


def test_no_crash_with_missing_gpos(tmpdir: Path):
    """Test that the optimize script gracefully handles TTFs with no GPOS."""

    # Create a test TTF.
    glyphs = ".notdef space A Aacute B D".split()
    fb = FontBuilder(1000)
    fb.setupGlyphOrder(glyphs)

    # Confirm that it has no GPOS.
    assert "GPOS" not in fb.font

    # Save, and feed to the optimize CLI.
    input = tmpdir / "in.ttf"
    fb.save(str(input))

    output = tmpdir / "out.ttf"
    args = [
        "--gpos-compression-level",
        "5",
        str(input),
        "-o",
        str(output),
    ]
    from fontTools.otlLib.optimize import main

    # Assert that we did not crash, and saved an output font.
    ret = main(args)
    assert ret in (0, None)
    assert output.exists()


# Copy-pasted from https://stackoverflow.com/questions/2059482/python-temporarily-modify-the-current-processs-environment
# TODO: remove when moving to the Config class
@contextlib.contextmanager
def set_env(**environ):
    """
    Temporarily set the process environment variables.

    >>> with set_env(PLUGINS_DIR=u'test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True

    >>> "PLUGINS_DIR" in os.environ
    False

    :type environ: dict[str, unicode]
    :param environ: Environment variables to set
    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


def count_pairpos_subtables(font: TTFont) -> int:
    subtables = 0
    for lookup in font["GPOS"].table.LookupList.Lookup:
        if lookup.LookupType == 2:
            subtables += len(lookup.SubTable)
        elif lookup.LookupType == 9:
            for subtable in lookup.SubTable:
                if subtable.ExtensionLookupType == 2:
                    subtables += 1
    return subtables


def count_pairpos_bytes(font: TTFont) -> int:
    bytes = 0
    gpos = font["GPOS"]
    for lookup in font["GPOS"].table.LookupList.Lookup:
        if lookup.LookupType == 2:
            w = OTTableWriter(tableTag=gpos.tableTag)
            lookup.compile(w, font)
            bytes += len(w.getAllData())
        elif lookup.LookupType == 9:
            if any(subtable.ExtensionLookupType == 2 for subtable in lookup.SubTable):
                w = OTTableWriter(tableTag=gpos.tableTag)
                lookup.compile(w, font)
                bytes += len(w.getAllData())
    return bytes


def get_kerning_by_blocks(blocks: List[Tuple[int, int]]) -> Tuple[List[str], str]:
    """Generate a highly compressible font by generating a bunch of rectangular
    blocks on the diagonal that can easily be sliced into subtables.

    Returns the list of glyphs and feature code of the font.
    """
    value = 0
    glyphs: List[str] = []
    rules = []
    # Each block is like a script in a multi-script font
    for script, (width, height) in enumerate(blocks):
        glyphs.extend(f"g_{script}_{i}" for i in range(max(width, height)))
        for l in range(height):
            for r in range(width):
                value += 1
                rules.append((f"g_{script}_{l}", f"g_{script}_{r}", value))
    classes = "\n".join([f"@{g} = [{g}];" for g in glyphs])
    statements = "\n".join([f"pos @{l} @{r} {v};" for (l, r, v) in rules])
    features = f"""
        {classes}
        feature kern {{
            {statements}
        }} kern;
    """
    return glyphs, features


@pytest.mark.parametrize(
    ("blocks", "level", "expected_subtables", "expected_bytes"),
    [
        # Level = 0 = no optimization leads to 650 bytes of GPOS
        ([(15, 3), (2, 10)], None, 1, 602),
        # Optimization level 1 recognizes the 2 blocks and splits into 2
        # subtables = adds 1 subtable leading to a size reduction of
        # (602-298)/602 = 50%
        ([(15, 3), (2, 10)], 1, 2, 298),
        # On a bigger block configuration, we see that mode=5 doesn't create
        # as many subtables as it could, because of the stop criteria
        ([(4, 4) for _ in range(20)], 5, 14, 2042),
        # while level=9 creates as many subtables as there were blocks on the
        # diagonal and yields a better saving
        ([(4, 4) for _ in range(20)], 9, 20, 1886),
        # On a fully occupied kerning matrix, even the strategy 9 doesn't
        # split anything.
        ([(10, 10)], 9, 1, 304),
    ],
)
def test_optimization_mode(
    caplog,
    blocks: List[Tuple[int, int]],
    level: Optional[int],
    expected_subtables: int,
    expected_bytes: int,
):
    """Check that the optimizations are off by default, and that increasing
    the optimization level creates more subtables and a smaller byte size.
    """
    caplog.set_level(logging.DEBUG)

    glyphs, features = get_kerning_by_blocks(blocks)
    glyphs = [".notdef space"] + glyphs

    fb = FontBuilder(1000)
    if level is not None:
        fb.font.cfg["fontTools.otlLib.optimize.gpos:COMPRESSION_LEVEL"] = level
    fb.setupGlyphOrder(glyphs)
    addOpenTypeFeaturesFromString(fb.font, features)
    assert expected_subtables == count_pairpos_subtables(fb.font)
    assert expected_bytes == count_pairpos_bytes(fb.font)



import io
import os
import re
from fontTools import ttLib
from fontTools.fontBuilder import FontBuilder
import unittest
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable, table__c_m_a_p

CURR_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR = os.path.join(CURR_DIR, "data")
CMAP_FORMAT_14_TTX = os.path.join(DATA_DIR, "_c_m_a_p_format_14.ttx")
CMAP_FORMAT_14_BW_COMPAT_TTX = os.path.join(
    DATA_DIR, "_c_m_a_p_format_14_bw_compat.ttx"
)


def strip_VariableItems(string):
    # ttlib changes with the fontTools version
    string = re.sub(' ttLibVersion=".*"', "", string)
    return string


class CmapSubtableTest(unittest.TestCase):
    def makeSubtable(self, cmapFormat, platformID, platEncID, langID):
        subtable = CmapSubtable.newSubtable(cmapFormat)
        subtable.platformID, subtable.platEncID, subtable.language = (
            platformID,
            platEncID,
            langID,
        )
        return subtable

    def test_toUnicode_utf16be(self):
        subtable = self.makeSubtable(4, 0, 2, 7)
        self.assertEqual("utf_16_be", subtable.getEncoding())
        self.assertEqual(True, subtable.isUnicode())

    def test_toUnicode_macroman(self):
        subtable = self.makeSubtable(4, 1, 0, 7)  # MacRoman
        self.assertEqual("mac_roman", subtable.getEncoding())
        self.assertEqual(False, subtable.isUnicode())

    def test_toUnicode_macromanian(self):
        subtable = self.makeSubtable(4, 1, 0, 37)  # Mac Romanian
        self.assertNotEqual(None, subtable.getEncoding())
        self.assertEqual(False, subtable.isUnicode())

    def test_extended_mac_encodings(self):
        subtable = self.makeSubtable(4, 1, 1, 0)  # Mac Japanese
        self.assertNotEqual(None, subtable.getEncoding())
        self.assertEqual(False, subtable.isUnicode())

    def test_extended_unknown(self):
        subtable = self.makeSubtable(4, 10, 11, 12)
        self.assertEqual(subtable.getEncoding(), None)
        self.assertEqual(subtable.getEncoding("ascii"), "ascii")
        self.assertEqual(subtable.getEncoding(default="xyz"), "xyz")

    def test_compile_2(self):
        subtable = self.makeSubtable(2, 1, 2, 0)
        subtable.cmap = {c: "cid%05d" % c for c in range(32, 8192)}
        font = ttLib.TTFont()
        font.setGlyphOrder([".notdef"] + list(subtable.cmap.values()))
        data = subtable.compile(font)

        subtable2 = CmapSubtable.newSubtable(2)
        subtable2.decompile(data, font)
        self.assertEqual(subtable2.cmap, subtable.cmap)

    def test_compile_2_rebuild_rev_glyph_order(self):
        for fmt in [2, 4, 12]:
            subtable = self.makeSubtable(fmt, 1, 2, 0)
            subtable.cmap = {c: "cid%05d" % c for c in range(32, 8192)}
            font = ttLib.TTFont()
            font.setGlyphOrder([".notdef"] + list(subtable.cmap.values()))
            font._reverseGlyphOrderDict = (
                {}
            )  # force first KeyError branch in subtable.compile()
            data = subtable.compile(font)
            subtable2 = CmapSubtable.newSubtable(fmt)
            subtable2.decompile(data, font)
            self.assertEqual(subtable2.cmap, subtable.cmap, str(fmt))

    def test_compile_2_gids(self):
        for fmt in [2, 4, 12]:
            subtable = self.makeSubtable(fmt, 1, 3, 0)
            subtable.cmap = {0x0041: "gid001", 0x0042: "gid002"}
            font = ttLib.TTFont()
            font.setGlyphOrder([".notdef"])
            data = subtable.compile(font)

    def test_compile_decompile_4_empty(self):
        subtable = self.makeSubtable(4, 3, 1, 0)
        subtable.cmap = {}
        font = ttLib.TTFont()
        font.setGlyphOrder([])
        data = subtable.compile(font)
        subtable2 = CmapSubtable.newSubtable(4)
        subtable2.decompile(data, font)
        self.assertEqual(subtable2.cmap, {})

    def test_decompile_4(self):
        subtable = CmapSubtable.newSubtable(4)
        font = ttLib.TTFont()
        font.setGlyphOrder([])
        subtable.decompile(b"\0" * 3 + b"\x10" + b"\0" * 12, font)

    def test_decompile_12(self):
        subtable = CmapSubtable.newSubtable(12)
        font = ttLib.TTFont()
        font.setGlyphOrder([])
        subtable.decompile(b"\0" * 7 + b"\x10" + b"\0" * 8, font)

    def test_buildReversed(self):
        c4 = self.makeSubtable(4, 3, 1, 0)
        c4.cmap = {0x0041: "A", 0x0391: "A"}
        c12 = self.makeSubtable(12, 3, 10, 0)
        c12.cmap = {0x10314: "u10314"}
        cmap = table__c_m_a_p()
        cmap.tables = [c4, c12]
        self.assertEqual(
            cmap.buildReversed(), {"A": {0x0041, 0x0391}, "u10314": {0x10314}}
        )

    def test_getBestCmap(self):
        c4 = self.makeSubtable(4, 3, 1, 0)
        c4.cmap = {0x0041: "A", 0x0391: "A"}
        c12 = self.makeSubtable(12, 3, 10, 0)
        c12.cmap = {0x10314: "u10314"}
        cmap = table__c_m_a_p()
        cmap.tables = [c4, c12]
        self.assertEqual(cmap.getBestCmap(), {0x10314: "u10314"})
        self.assertEqual(
            cmap.getBestCmap(cmapPreferences=[(3, 1)]), {0x0041: "A", 0x0391: "A"}
        )
        self.assertEqual(cmap.getBestCmap(cmapPreferences=[(0, 4)]), None)

    def test_font_getBestCmap(self):
        c4 = self.makeSubtable(4, 3, 1, 0)
        c4.cmap = {0x0041: "A", 0x0391: "A"}
        c12 = self.makeSubtable(12, 3, 10, 0)
        c12.cmap = {0x10314: "u10314"}
        cmap = table__c_m_a_p()
        cmap.tables = [c4, c12]
        font = ttLib.TTFont()
        font["cmap"] = cmap
        self.assertEqual(font.getBestCmap(), {0x10314: "u10314"})
        self.assertEqual(
            font.getBestCmap(cmapPreferences=[(3, 1)]), {0x0041: "A", 0x0391: "A"}
        )
        self.assertEqual(font.getBestCmap(cmapPreferences=[(0, 4)]), None)

    def test_format_14(self):
        subtable = self.makeSubtable(14, 0, 5, 0)
        subtable.cmap = {}  # dummy
        subtable.uvsDict = {
            0xFE00: [(0x0030, "zero.slash")],
            0xFE01: [(0x0030, None)],
        }
        fb = FontBuilder(1024, isTTF=True)
        font = fb.font
        fb.setupGlyphOrder([".notdef", "zero.slash"])
        fb.setupMaxp()
        fb.setupPost()
        cmap = table__c_m_a_p()
        cmap.tableVersion = 0
        cmap.tables = [subtable]
        font["cmap"] = cmap
        f = io.BytesIO()
        font.save(f)
        f.seek(0)
        font = ttLib.TTFont(f)
        self.assertEqual(font["cmap"].getcmap(0, 5).uvsDict, subtable.uvsDict)
        f = io.StringIO(newline=None)
        font.saveXML(f, tables=["cmap"])
        ttx = strip_VariableItems(f.getvalue())
        with open(CMAP_FORMAT_14_TTX) as f:
            expected = strip_VariableItems(f.read())
        self.assertEqual(ttx, expected)
        with open(CMAP_FORMAT_14_BW_COMPAT_TTX) as f:
            font.importXML(f)
        self.assertEqual(font["cmap"].getcmap(0, 5).uvsDict, subtable.uvsDict)

    def test_sort_subtables_with_duplicate_keys(self):
        # https://github.com/fonttools/fonttools/issues/4035
        # Sorting subtables that share (platformID, platEncID, language) but
        # differ in format should not raise TypeError from comparing dicts.
        subtable4 = self.makeSubtable(4, 3, 1, 0)
        subtable4.cmap = {0x41: "A"}
        subtable12 = self.makeSubtable(12, 3, 1, 0)
        subtable12.cmap = {0x41: "A"}
        # This used to raise: TypeError: '<' not supported between
        # instances of 'dict' and 'dict'
        tables = sorted([subtable12, subtable4])
        # Both have the same key, so stable sort preserves original order
        self.assertEqual(len(tables), 2)

    def test_compile_raises_on_duplicate_subtable_keys(self):
        # https://github.com/fonttools/fonttools/issues/4035
        # The OpenType spec requires each (platformID, platEncID, language)
        # combination to be unique; compile should raise a clear error.
        cmap = table__c_m_a_p()
        cmap.tableVersion = 0
        subtable4 = self.makeSubtable(4, 3, 1, 0)
        subtable4.cmap = {0x41: "A"}
        subtable12 = self.makeSubtable(12, 3, 1, 0)
        subtable12.cmap = {0x41: "A"}
        cmap.tables = [subtable4, subtable12]
        with self.assertRaisesRegex(ValueError, "duplicate"):
            cmap.compile(ttFont=None)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from io import StringIO
from fontTools.designspaceLib import (
    AxisDescriptor,
    DesignSpaceDocument,
    SourceDescriptor,
)
from fontTools.fontBuilder import FontBuilder
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.varLib import build
from fontTools.varLib.builder import buildVarData
import pytest


@pytest.mark.parametrize(
    "region_indices, items, expected_num_shorts",
    [
        ([], [], 0),
        ([0], [[1]], 0),
        ([0], [[128]], 1),
        ([0, 1, 2], [[128, 1, 2], [3, -129, 5], [6, 7, 8]], 2),
        ([0, 1, 2], [[0, 128, 2], [3, 4, 5], [6, 7, -129]], 3),
        ([0], [[32768]], 0x8001),
        ([0, 1, 2], [[32768, 1, 2], [3, -129, 5], [6, 7, 8]], 0x8001),
        ([0, 1, 2], [[32768, 1, 2], [3, -32769, 5], [6, 7, 8]], 0x8002),
        ([0, 1, 2], [[0, 32768, 2], [3, 4, 5], [6, 7, -32769]], 0x8003),
    ],
    ids=[
        "0_regions_0_deltas",
        "1_region_1_uint8",
        "1_region_1_short",
        "3_regions_2_shorts_ordered",
        "3_regions_2_shorts_unordered",
        "1_region_1_long",
        "3_regions_1_long_ordered",
        "3_regions_2_longs_ordered",
        "3_regions_2_longs_unordered",
    ],
)
def test_buildVarData_no_optimize(region_indices, items, expected_num_shorts):
    data = buildVarData(region_indices, items, optimize=False)

    assert data.ItemCount == len(items)
    assert data.NumShorts == expected_num_shorts
    assert data.VarRegionCount == len(region_indices)
    assert data.VarRegionIndex == region_indices
    assert data.Item == items


@pytest.mark.parametrize(
    [
        "region_indices",
        "items",
        "expected_num_shorts",
        "expected_regions",
        "expected_items",
    ],
    [
        (
            [0, 1, 2],
            [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
            0,
            [0, 1, 2],
            [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 128, 2], [3, 4, 5], [6, 7, 8]],
            1,
            [1, 0, 2],
            [[128, 0, 2], [4, 3, 5], [7, 6, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 1, 128], [3, 4, 5], [6, -129, 8]],
            2,
            [1, 2, 0],
            [[1, 128, 0], [4, 5, 3], [-129, 8, 6]],
        ),
        (
            [0, 1, 2],
            [[128, 1, -129], [3, 4, 5], [6, 7, 8]],
            2,
            [0, 2, 1],
            [[128, -129, 1], [3, 5, 4], [6, 8, 7]],
        ),
        (
            [0, 1, 2],
            [[0, 1, 128], [3, -129, 5], [256, 7, 8]],
            3,
            [0, 1, 2],
            [[0, 1, 128], [3, -129, 5], [256, 7, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 128, 2], [0, 4, 5], [0, 7, 8]],
            1,
            [1, 2],
            [[128, 2], [4, 5], [7, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 32768, 2], [3, 4, 5], [6, 7, 8]],
            0x8001,
            [1, 0, 2],
            [[32768, 0, 2], [4, 3, 5], [7, 6, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 1, 32768], [3, 4, 5], [6, -32769, 8]],
            0x8002,
            [1, 2, 0],
            [[1, 32768, 0], [4, 5, 3], [-32769, 8, 6]],
        ),
        (
            [0, 1, 2],
            [[32768, 1, -32769], [3, 4, 5], [6, 7, 8]],
            0x8002,
            [0, 2, 1],
            [[32768, -32769, 1], [3, 5, 4], [6, 8, 7]],
        ),
        (
            [0, 1, 2],
            [[0, 1, 32768], [3, -32769, 5], [65536, 7, 8]],
            0x8003,
            [0, 1, 2],
            [[0, 1, 32768], [3, -32769, 5], [65536, 7, 8]],
        ),
        (
            [0, 1, 2],
            [[0, 32768, 2], [0, 4, 5], [0, 7, 8]],
            0x8001,
            [1, 2],
            [[32768, 2], [4, 5], [7, 8]],
        ),
    ],
    ids=[
        "0/3_shorts_no_reorder",
        "1/3_shorts_reorder",
        "2/3_shorts_reorder",
        "2/3_shorts_same_row_reorder",
        "3/3_shorts_no_reorder",
        "1/3_shorts_1/3_zeroes",
        "1/3_longs_reorder",
        "2/3_longs_reorder",
        "2/3_longs_same_row_reorder",
        "3/3_longs_no_reorder",
        "1/3_longs_1/3_zeroes",
    ],
)
def test_buildVarData_optimize(
    region_indices, items, expected_num_shorts, expected_regions, expected_items
):
    data = buildVarData(region_indices, items, optimize=True)

    assert data.ItemCount == len(items)
    assert data.NumShorts == expected_num_shorts
    assert data.VarRegionCount == len(expected_regions)
    assert data.VarRegionIndex == expected_regions
    assert data.Item == expected_items


def test_empty_vhvar_size():
    """HVAR/VHVAR should be present but empty when there are no glyph metrics
    variations, and should use a direct mapping for optimal encoding."""

    # Make a designspace that varies the outlines of 'A' but not its advance.
    doc = DesignSpaceDocument()

    doc.addAxis(
        AxisDescriptor(tag="wght", name="Weight", minimum=400, default=400, maximum=700)
    )

    for wght in (400, 700):
        # Outlines depend on weight.
        pen = TTGlyphPen(None)
        pen.lineTo((0, wght))
        pen.lineTo((wght, wght))
        pen.lineTo((wght, 0))
        pen.closePath()
        glyphs = {"A": pen.glyph()}

        fb = FontBuilder(unitsPerEm=1000)
        fb.setupGlyphOrder(list(glyphs.keys()))
        fb.setupGlyf(glyphs)

        # Horizontal advance does not vary.
        fb.setupHorizontalMetrics(
            {name: (500, fb.font["glyf"][name].xMin) for name in glyphs}  # type: ignore
        )
        fb.setupHorizontalHeader(ascent=1000, descent=0)

        # Vertical advance does not vary.
        fb.setupVerticalMetrics(
            {name: (500, 1000 - fb.font["glyf"][name].yMax) for name in glyphs}  # type: ignore
        )
        fb.setupVerticalHeader(ascent=1000, descent=0)

        fb.setupNameTable({"familyName": "TestEmptyVhvar", "styleName": "Regular"})
        fb.setupPost()
        doc.addSource(SourceDescriptor(font=fb.font, location={"Weight": wght}))

    # Compile.
    vf, *_ = build(doc)

    # Test both tables' encodings:
    for table in ("HVAR", "VVAR"):
        # Variable glyph metrics table should be built even when there are no
        # glyph metrics variations.
        assert table in vf

        # The table should be empty, and use a direct mapping for optimal size.
        expected = """
<?xml version="1.0" encoding="UTF-8"?>
<Version value="0x00010000"/>
<VarStore Format="1">
  <Format value="1"/>
  <VarRegionList>
    <!-- RegionAxisCount=1 -->
    <!-- RegionCount=0 -->
  </VarRegionList>
  <!-- VarDataCount=1 -->
  <VarData index="0">
    <!-- ItemCount=1 -->
    <NumShorts value="0"/>
    <!-- VarRegionCount=0 -->
    <Item index="0" value="[]"/>
  </VarData>
</VarStore>
""".lstrip()

        with StringIO() as buffer:
            writer = XMLWriter(buffer)
            vf[table].toXML(writer, vf)
            actual = buffer.getvalue()
        assert actual == expected

        # The table should be encodable in at least this size.
        # (VVAR has an extra Offset32 to point to a vertical origin mapping)
        optimal_size = {"HVAR": 42, "VVAR": 46}[table]
        assert len(vf[table].compile(vf)) <= optimal_size


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



import unittest
import os
import math

try:
    from fontTools.pens.freetypePen import FreeTypePen

    FREETYPE_PY_AVAILABLE = True
except ImportError:
    FREETYPE_PY_AVAILABLE = False

from fontTools.misc.transform import Scale, Offset

DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data")


def box(pen, offset=(0, 0)):
    pen.moveTo((0 + offset[0], 0 + offset[1]))
    pen.lineTo((0 + offset[0], 500 + offset[1]))
    pen.lineTo((500 + offset[0], 500 + offset[1]))
    pen.lineTo((500 + offset[0], 0 + offset[1]))
    pen.closePath()


def draw_cubic(pen):
    pen.moveTo((50, 0))
    pen.lineTo((50, 500))
    pen.lineTo((200, 500))
    pen.curveTo((350, 500), (450, 400), (450, 250))
    pen.curveTo((450, 100), (350, 0), (200, 0))
    pen.closePath()


def draw_quadratic(pen):
    pen.moveTo((50, 0))
    pen.lineTo((50, 500))
    pen.lineTo((200, 500))
    pen.qCurveTo((274, 500), (388, 438), (450, 324), (450, 250))
    pen.qCurveTo((450, 176), (388, 62), (274, 0), (200, 0))
    pen.closePath()


def star(pen):
    pen.moveTo((0, 420))
    pen.lineTo((1000, 420))
    pen.lineTo((200, -200))
    pen.lineTo((500, 800))
    pen.lineTo((800, -200))
    pen.closePath()


# For the PGM format, see the following resources:
#   https://en.wikipedia.org/wiki/Netpbm
#   http://netpbm.sourceforge.net/doc/pgm.html
def load_pgm(filename):
    with open(filename, "rb") as fp:
        assert fp.readline() == "P5\n".encode()
        w, h = (int(c) for c in fp.readline().decode().rstrip().split(" "))
        assert fp.readline() == "255\n".encode()
        return fp.read(), (w, h)


def save_pgm(filename, buf, size):
    with open(filename, "wb") as fp:
        fp.write("P5\n".encode())
        fp.write("{0:d} {1:d}\n".format(*size).encode())
        fp.write("255\n".encode())
        fp.write(buf)


# Assume the buffers are equal when PSNR > 38dB. See also:
#   Peak signal-to-noise ratio
#   https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
PSNR_THRESHOLD = 38.0


def psnr(b1, b2):
    import math

    mse = sum((c1 - c2) * (c1 - c2) for c1, c2 in zip(b1, b2)) / float(len(b1))
    return 10.0 * math.log10((255.0**2) / float(mse)) if mse > 0 else math.inf


@unittest.skipUnless(FREETYPE_PY_AVAILABLE, "freetype-py not installed")
class FreeTypePenTest(unittest.TestCase):
    def test_draw(self):
        pen = FreeTypePen(None)
        box(pen)
        width, height = 500, 500
        buf1, _ = pen.buffer(width=width, height=height)
        buf2 = b"\xff" * width * height
        self.assertEqual(buf1, buf2)

    def test_empty(self):
        pen = FreeTypePen(None)
        width, height = 500, 500
        buf, size = pen.buffer(width=width, height=height)
        self.assertEqual(b"\0" * size[0] * size[1], buf)

    def test_bbox_and_cbox(self):
        pen = FreeTypePen(None)
        draw_cubic(pen)
        self.assertEqual(pen.bbox, (50.0, 0.0, 450.0, 500.0))
        self.assertEqual(pen.cbox, (50.0, 0.0, 450.0, 500.0))

    def test_non_zero_fill(self):
        pen = FreeTypePen(None)
        star(pen)
        t = Scale(0.05, 0.05)
        width, height = t.transformPoint((1000, 1000))
        t = t.translate(0, 200)
        buf1, size1 = pen.buffer(width=width, height=height, transform=t, evenOdd=False)
        buf2, size2 = load_pgm(os.path.join(DATA_DIR, "test_non_zero_fill.pgm"))
        self.assertEqual(len(buf1), len(buf2))
        self.assertEqual(size1, size2)
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_even_odd_fill(self):
        pen = FreeTypePen(None)
        star(pen)
        t = Scale(0.05, 0.05)
        width, height = t.transformPoint((1000, 1000))
        t = t.translate(0, 200)
        buf1, size1 = pen.buffer(width=width, height=height, transform=t, evenOdd=True)
        buf2, size2 = load_pgm(os.path.join(DATA_DIR, "test_even_odd_fill.pgm"))
        self.assertEqual(len(buf1), len(buf2))
        self.assertEqual(size1, size2)
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_cubic_vs_quadratic(self):
        pen1, pen2 = FreeTypePen(None), FreeTypePen(None)
        draw_cubic(pen1)
        draw_quadratic(pen2)
        width, height = 500, 500
        buf1, _ = pen1.buffer(width=width, height=height)
        buf2, _ = pen2.buffer(width=width, height=height)
        self.assertEqual(len(buf1), len(buf2))
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_contain(self):
        pen = FreeTypePen(None)
        star(pen)
        t = Scale(0.05, 0.05)
        width, height = 0, 0
        buf1, size1 = pen.buffer(width=width, height=height, transform=t, contain=True)
        buf2, size2 = load_pgm(os.path.join(DATA_DIR, "test_non_zero_fill.pgm"))
        self.assertEqual(len(buf1), len(buf2))
        self.assertEqual(size1, size2)
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_rotate(self):
        pen = FreeTypePen(None)
        box(pen)
        t = Scale(0.05, 0.05).rotate(math.pi / 4.0).translate(1234, 5678)
        width, height = None, None
        buf1, size1 = pen.buffer(width=width, height=height, transform=t)
        buf2, size2 = load_pgm(os.path.join(DATA_DIR, "test_rotate.pgm"))
        self.assertEqual(len(buf1), len(buf2))
        self.assertEqual(size1, size2)
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_skew(self):
        pen = FreeTypePen(None)
        box(pen)
        t = Scale(0.05, 0.05).skew(math.pi / 4.0).translate(1234, 5678)
        width, height = None, None
        buf1, size1 = pen.buffer(width=width, height=height, transform=t)
        buf2, size2 = load_pgm(os.path.join(DATA_DIR, "test_skew.pgm"))
        self.assertEqual(len(buf1), len(buf2))
        self.assertEqual(size1, size2)
        self.assertGreater(psnr(buf1, buf2), PSNR_THRESHOLD)

    def test_none_size(self):
        pen = FreeTypePen(None)
        star(pen)
        width, height = None, None
        buf1, size = pen.buffer(width=width, height=height, transform=Offset(0, 200))
        buf2, _ = pen.buffer(width=1000, height=1000, transform=Offset(0, 200))
        self.assertEqual(size, (1000, 1000))
        self.assertEqual(buf1, buf2)

        pen = FreeTypePen(None)
        box(pen, offset=(250, 250))
        width, height = None, None
        buf1, size = pen.buffer(width=width, height=height)
        buf2, _ = pen.buffer(width=500, height=500, transform=Offset(-250, -250))
        self.assertEqual(size, (500, 500))
        self.assertEqual(buf1, buf2)

        pen = FreeTypePen(None)
        box(pen, offset=(-1234, -5678))
        width, height = None, None
        buf1, size = pen.buffer(width=width, height=height)
        buf2, _ = pen.buffer(width=500, height=500, transform=Offset(1234, 5678))
        self.assertEqual(size, (500, 500))
        self.assertEqual(buf1, buf2)

    def test_zero_size(self):
        pen = FreeTypePen(None)
        star(pen)
        width, height = 0, 0
        buf1, size = pen.buffer(
            width=width, height=height, transform=Offset(0, 200), contain=True
        )
        buf2, _ = pen.buffer(
            width=1000, height=1000, transform=Offset(0, 200), contain=True
        )
        self.assertEqual(size, (1000, 1000))
        self.assertEqual(buf1, buf2)

        pen = FreeTypePen(None)
        box(pen, offset=(250, 250))
        width, height = 0, 0
        buf1, size = pen.buffer(width=width, height=height, contain=True)
        buf2, _ = pen.buffer(
            width=500, height=500, transform=Offset(0, 0), contain=True
        )
        self.assertEqual(size, (750, 750))
        self.assertEqual(buf1, buf2)

        pen = FreeTypePen(None)
        box(pen, offset=(-1234, -5678))
        width, height = 0, 0
        buf1, size = pen.buffer(width=width, height=height, contain=True)
        buf2, _ = pen.buffer(
            width=500, height=500, transform=Offset(1234, 5678), contain=True
        )
        self.assertEqual(size, (500, 500))
        self.assertEqual(buf1, buf2)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import os
from pathlib import Path
from fontTools.ttLib import TTCollection
import pytest

TTX_DATA_DIR = Path(__file__).parent.parent / "ttx" / "data"


@pytest.fixture(params=[None, True, False])
def lazy(request):
    return request.param


def test_lazy_open_path(lazy):
    ttc_path = TTX_DATA_DIR / "TestTTC.ttc"
    with TTCollection(ttc_path, lazy=lazy) as collection:
        assert len(collection) == 2
        assert collection[0]["maxp"].numGlyphs == 6
        assert collection[1]["maxp"].numGlyphs == 6


def test_lazy_open_file(lazy):
    with (TTX_DATA_DIR / "TestTTC.ttc").open("rb") as file:
        collection = TTCollection(file, lazy=lazy)
        assert len(collection) == 2
        assert collection[0]["maxp"].numGlyphs == 6
        assert collection[1]["maxp"].numGlyphs == 6
